import gzip
import os

from shutil import copyfileobj
from warnings import warn

from igraph._igraph import GraphBase
from igraph.utils import (
    named_temporary_file,
)


def _identify_format(filename):
    """_identify_format(filename)

    Tries to identify the format of the graph stored in the file with the
    given filename. It identifies most file formats based on the extension
    of the file (and not on syntactic evaluation). The only exception is
    the adjacency matrix format and the edge list format: the first few
    lines of the file are evaluated to decide between the two.

    @note: Internal function, should not be called directly.

    @param filename: the name of the file or a file object whose C{name}
      attribute is set.
    @return: the format of the file as a string.
    """
    import os.path

    if hasattr(filename, "name") and hasattr(filename, "read"):
        # It is most likely a file
        try:
            filename = filename.name
        except Exception:
            return None

    root, ext = os.path.splitext(filename)
    ext = ext.lower()

    if ext == ".gz":
        _, ext2 = os.path.splitext(root)
        ext2 = ext2.lower()
        if ext2 == ".pickle":
            return "picklez"
        elif ext2 == ".graphml":
            return "graphmlz"

    if ext in [
        ".dimacs",
        ".dl",
        ".dot",
        ".edge",
        ".edges",
        ".edgelist",
        ".gml",
        ".graphml",
        ".graphmlz",
        ".gw",
        ".lgl",
        ".lgr",
        ".ncol",
        ".net",
        ".pajek",
        ".pickle",
        ".picklez",
        ".svg",
    ]:
        return ext[1:]

    if ext == ".txt" or ext == ".dat":
        # Most probably an adjacency matrix or an edge list
        f = open(filename, "r")
        line = f.readline()
        if line is None:
            return "edges"
        parts = line.strip().split()
        if len(parts) == 2:
            line = f.readline()
            if line is None:
                return "edges"
            parts = line.strip().split()
            if len(parts) == 2:
                line = f.readline()
                if line is None:
                    # This is a 2x2 matrix, it can be a matrix or an edge
                    # list as well and we cannot decide
                    return None
                else:
                    parts = line.strip().split()
                    if len(parts) == 0:
                        return None
                return "edges"
            else:
                # Not a matrix
                return None
        else:
            return "adjacency"


def _construct_graph_from_adjacency_file(
    cls, f, sep=None, comment_char="#", attribute=None, *args, **kwds
):
    """Constructs a graph based on an adjacency matrix from the given file.

    Additional positional and keyword arguments not mentioned here are
    passed intact to L{Graph.Adjacency}.

    @param f: the name of the file to be read or a file object
    @param sep: the string that separates the matrix elements in a row.
      C{None} means an arbitrary sequence of whitespace characters.
    @param comment_char: lines starting with this string are treated
      as comments.
    @param attribute: an edge attribute name where the edge weights are
      stored in the case of a weighted adjacency matrix. If C{None},
      no weights are stored, values larger than 1 are considered as
      edge multiplicities.
    @return: the created graph"""
    if isinstance(f, str):
        f = open(f)

    matrix, ri = [], 0
    for line in f:
        line = line.strip()
        if len(line) == 0:
            continue
        if line.startswith(comment_char):
            continue
        row = [float(x) for x in line.split(sep)]
        matrix.append(row)
        ri += 1

    f.close()

    if attribute is None:
        graph = cls.Adjacency(matrix, *args, **kwds)
    else:
        graph, weights = cls._Weighted_Adjacency(matrix, *args, **kwds)
        graph.es[attribute] = weights

    return graph


def _construct_graph_from_dimacs_file(cls, f, directed=False):
    """Reads a graph from a file conforming to the DIMACS minimum-cost flow
    file format.

    For the exact definition of the format, see
    U{http://lpsolve.sourceforge.net/5.5/DIMACS.htm}.

    Restrictions compared to the official description of the format are
    as follows:

      - igraph's DIMACS reader requires only three fields in an arc
        definition, describing the edge's source and target node and
        its capacity.
      - Source vertices are identified by 's' in the FLOW field, target
        vertices are identified by 't'.
      - Node indices start from 1. Only a single source and target node
        is allowed.

    @param f: the name of the file or a Python file handle
    @param directed: whether the generated graph should be directed.
    @return: the generated graph. The indices of the source and target
      vertices are attached as graph attributes C{source} and C{target},
      the edge capacities are stored in the C{capacity} edge attribute.
    """
    # Deferred import to avoid cycles
    from igraph import Graph

    graph, source, target, cap = super(Graph, cls).Read_DIMACS(f, directed)
    graph.es["capacity"] = cap
    graph["source"] = source
    graph["target"] = target
    return graph


def _construct_graph_from_graphmlz_file(cls, f, index=0):
    """Reads a graph from a zipped GraphML file.

    @param f: the name of the file
    @param index: if the GraphML file contains multiple graphs,
      specified the one that should be loaded. Graph indices
      start from zero, so if you want to load the first graph,
      specify 0 here.
    @return: the loaded graph object"""
    with named_temporary_file() as tmpfile:
        with open(tmpfile, "wb") as outf:
            copyfileobj(gzip.GzipFile(f, "rb"), outf)
        return cls.Read_GraphML(tmpfile, index=index)


def _construct_graph_from_pickle_file(cls, fname=None):
    """Reads a graph from Python pickled format

    @param fname: the name of the file, a stream to read from, or
      a string containing the pickled data.
    @return: the created graph object.
    """
    import pickle as pickle

    if hasattr(fname, "read"):
        # Probably a file or a file-like object
        result = pickle.load(fname)
    else:
        try:
            fp = open(fname, "rb")
        except UnicodeDecodeError:
            try:
                # We are on Python 3.6 or above and we are passing a pickled
                # stream that cannot be decoded as Unicode. Try unpickling
                # directly.
                result = pickle.loads(fname)
            except TypeError:
                raise IOError(
                    "Cannot load file. If fname is a file name, that "
                    "filename may be incorrect."
                ) from None
        except IOError:
            try:
                # No file with the given name, try unpickling directly.
                result = pickle.loads(fname)
            except TypeError:
                raise IOError(
                    "Cannot load file. If fname is a file name, that "
                    "filename may be incorrect."
                ) from None
        else:
            result = pickle.load(fp)
            fp.close()

    if not isinstance(result, cls):
        raise TypeError("unpickled object is not a %s" % cls.__name__)

    return result


def _construct_graph_from_picklez_file(cls, fname):
    """Reads a graph from compressed Python pickled format, uncompressing
    it on-the-fly.

    @param fname: the name of the file or a stream to read from.
    @return: the created graph object.
    """
    import pickle as pickle

    if hasattr(fname, "read"):
        # Probably a file or a file-like object
        if isinstance(fname, gzip.GzipFile):
            result = pickle.load(fname)
        else:
            result = pickle.load(gzip.GzipFile(mode="rb", fileobj=fname))
    else:
        result = pickle.load(gzip.open(fname, "rb"))

    if not isinstance(result, cls):
        raise TypeError("unpickled object is not a %s" % cls.__name__)

    return result


def _construct_graph_from_file(cls, f, format=None, *args, **kwds):
    """Unified reading function for graphs.

    This method tries to identify the format of the graph given in
    the first parameter and calls the corresponding reader method.

    The remaining arguments are passed to the reader method without
    any changes.

    @param f: the file containing the graph to be loaded
    @param format: the format of the file (if known in advance).
      C{None} means auto-detection. Possible values are: C{"ncol"}
      (NCOL format), C{"lgl"} (LGL format), C{"graphdb"} (GraphDB
      format), C{"graphml"}, C{"graphmlz"} (GraphML and gzipped
      GraphML format), C{"gml"} (GML format), C{"net"}, C{"pajek"}
      (Pajek format), C{"dimacs"} (DIMACS format), C{"edgelist"},
      C{"edges"} or C{"edge"} (edge list), C{"adjacency"}
      (adjacency matrix), C{"dl"} (DL format used by UCINET),
      C{"pickle"} (Python pickled format),
      C{"picklez"} (gzipped Python pickled format)
    @raises IOError: if the file format can't be identified and
      none was given.
    """
    if isinstance(f, os.PathLike):
        f = str(f)
    if format is None:
        format = _identify_format(f)
    try:
        reader = cls._format_mapping[format][0]
    except (KeyError, IndexError):
        raise IOError("unknown file format: %s" % str(format)) from None
    if reader is None:
        raise IOError("no reader method for file format: %s" % str(format))
    reader = getattr(cls, reader)
    return reader(f, *args, **kwds)


def _write_graph_to_adjacency_file(graph, f, sep=" ", eol="\n", *args, **kwds):
    """Writes the adjacency matrix of the graph to the given file

    All the remaining arguments not mentioned here are passed intact
    to L{Graph.get_adjacency}.

    @param f: the name of the file to be written.
    @param sep: the string that separates the matrix elements in a row
    @param eol: the string that separates the rows of the matrix. Please
      note that igraph is able to read back the written adjacency matrix
      if and only if this is a single newline character
    """
    if isinstance(f, str):
        f = open(f, "w")
    matrix = graph.get_adjacency(*args, **kwds)
    for row in matrix:
        f.write(sep.join(map(str, row)))
        f.write(eol)
    f.close()


def _write_graph_to_dimacs_file(
    graph, f, source=None, target=None, capacity="capacity"
):
    """Writes the graph in DIMACS format to the given file.

    @param f: the name of the file to be written or a Python file handle.
    @param source: the source vertex ID. If C{None}, igraph will try to
      infer it from the C{source} graph attribute.
    @param target: the target vertex ID. If C{None}, igraph will try to
      infer it from the C{target} graph attribute.
    @param capacity: the capacities of the edges in a list or the name of
      an edge attribute that holds the capacities. If there is no such
      edge attribute, every edge will have a capacity of 1.
    """
    if source is None:
        try:
            source = graph["source"]
        except KeyError:
            raise ValueError(
                "source vertex must be provided in the 'source' graph "
                "attribute or in the 'source' argument of write_dimacs()"
            ) from None

    if target is None:
        try:
            target = graph["target"]
        except KeyError:
            raise ValueError(
                "target vertex must be provided in the 'target' graph "
                "attribute or in the 'target' argument of write_dimacs()"
            ) from None

    if isinstance(capacity, str) and capacity not in graph.edge_attributes():
        warn("'%s' edge attribute does not exist" % capacity, stacklevel=1)
        capacity = [1] * graph.ecount()

    return GraphBase.write_dimacs(graph, f, source, target, capacity)


def _write_graph_to_graphmlz_file(graph, f, compresslevel=9):
    """Writes the graph to a zipped GraphML file.

    The library uses the gzip compression algorithm, so the resulting
    file can be unzipped with regular gzip uncompression (like
    C{gunzip} or C{zcat} from Unix command line) or the Python C{gzip}
    module.

    Uses a temporary file to store intermediate GraphML data, so
    make sure you have enough free space to store the unzipped
    GraphML file as well.

    @param f: the name of the file to be written.
    @param compresslevel: the level of compression. 1 is fastest and
      produces the least compression, and 9 is slowest and produces
      the most compression."""
    with named_temporary_file() as tmpfile:
        graph.write_graphml(tmpfile)
        outf = gzip.GzipFile(f, "wb", compresslevel)
        copyfileobj(open(tmpfile, "rb"), outf)
        outf.close()


def _write_graph_to_pickle_file(graph, fname=None, version=-1):
    """Saves the graph in Python pickled format

    @param fname: the name of the file or a stream to save to. If
      C{None}, saves the graph to a string and returns the string.
    @param version: pickle protocol version to be used. If -1, uses
      the highest protocol available
    @return: C{None} if the graph was saved successfully to the
      given file, or a string if C{fname} was C{None}.
    """
    import pickle as pickle

    if fname is None:
        return pickle.dumps(graph, version)
    if not hasattr(fname, "write"):
        file_was_opened = True
        fname = open(fname, "wb")
    else:
        file_was_opened = False
    result = pickle.dump(graph, fname, version)
    if file_was_opened:
        fname.close()
    return result


def _write_graph_to_picklez_file(graph, fname=None, version=-1):
    """Saves the graph in Python pickled format, compressed with
    gzip.

    Saving in this format is a bit slower than saving in a Python pickle
    without compression, but the final file takes up much less space on
    the hard drive.

    @param fname: the name of the file or a stream to save to.
    @param version: pickle protocol version to be used. If -1, uses
      the highest protocol available
    @return: C{None} if the graph was saved successfully to the
      given file.
    """
    import pickle as pickle

    file_was_opened = False

    if not hasattr(fname, "write"):
        file_was_opened = True
        fname = gzip.open(fname, "wb")
    elif not isinstance(fname, gzip.GzipFile):
        file_was_opened = True
        fname = gzip.GzipFile(mode="wb", fileobj=fname)

    result = pickle.dump(graph, fname, version)

    if file_was_opened:
        fname.close()

    return result


def _write_graph_to_file(graph, f, format=None, *args, **kwds):
    """Unified writing function for graphs.

    This method tries to identify the format of the graph given in
    the first parameter (based on extension) and calls the corresponding
    writer method.

    The remaining arguments are passed to the writer method without
    any changes.

    @param f: the file containing the graph to be saved
    @param format: the format of the file (if one wants to override the
      format determined from the filename extension, or the filename itself
      is a stream). C{None} means auto-detection. Possible values are:

        - C{"adjacency"}: adjacency matrix format

        - C{"dimacs"}: DIMACS format

        - C{"dot"}, C{"graphviz"}: GraphViz DOT format

        - C{"edgelist"}, C{"edges"} or C{"edge"}: numeric edge list format

        - C{"gml"}: GML format

        - C{"graphml"} and C{"graphmlz"}: standard and gzipped GraphML
          format

        - C{"gw"}, C{"leda"}, C{"lgr"}: LEDA native format

        - C{"lgl"}: LGL format

        - C{"ncol"}: NCOL format

        - C{"net"}, C{"pajek"}: Pajek format

        - C{"pickle"}, C{"picklez"}: standard and gzipped Python pickled
          format

        - C{"svg"}: SVG format

    @raises IOError: if the file format can't be identified and
      none was given.
    """
    if isinstance(f, os.PathLike):
        f = str(f)
    if format is None:
        format = _identify_format(f)
    try:
        writer = graph._format_mapping[format][1]
    except (KeyError, IndexError):
        raise IOError("unknown file format: %s" % str(format)) from None
    if writer is None:
        raise IOError("no writer method for file format: %s" % str(format))
    writer = getattr(graph, writer)
    return writer(f, *args, **kwds)



# vim:ts=4:sw=4:sts=4:et
# -*- coding: utf-8 -*-
"""Classes related to graph clustering."""

from copy import deepcopy
from io import StringIO

from igraph._igraph import GraphBase, community_to_membership, _compare_communities
from igraph.configuration import Configuration
from igraph.datatypes import UniqueIdGenerator
from igraph.drawing.colors import ClusterColoringPalette
from igraph.drawing.cairo.dendrogram import CairoDendrogramDrawer
from igraph.drawing.matplotlib.dendrogram import MatplotlibDendrogramDrawer
from igraph.statistics import Histogram
from igraph.summary import _get_wrapper_for_width
from igraph.utils import deprecated


class Clustering:
    """Class representing a clustering of an arbitrary ordered set.

    This is now used as a base for L{VertexClustering}, but it might be
    useful for other purposes as well.

    Members of an individual cluster can be accessed by the C{[]} operator:

      >>> cl = Clustering([0,0,0,0,1,1,1,2,2,2,2])
      >>> cl[0]
      [0, 1, 2, 3]

    The membership vector can be accessed by the C{membership} property:

      >>> cl.membership
      [0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2]

    The number of clusters can be retrieved by the C{len} function:

      >>> len(cl)
      3

    You can iterate over the clustering object as if it were a regular list
    of clusters:

      >>> for cluster in cl:
      ...     print(" ".join(str(idx) for idx in cluster))
      ...
      0 1 2 3
      4 5 6
      7 8 9 10

    If you need all the clusters at once as lists, you can simply convert
    the clustering object to a list:

      >>> cluster_list = list(cl)
      >>> print(cluster_list)
      [[0, 1, 2, 3], [4, 5, 6], [7, 8, 9, 10]]
    """

    def __init__(self, membership, params=None):
        """Constructor.

        @param membership: the membership list -- that is, the cluster
          index in which each element of the set belongs to.
        @param params: additional parameters to be stored in this
          object's dictionary.
        """
        self._membership = list(membership)
        if len(self._membership) > 0:
            self._len = max(m for m in self._membership if m is not None) + 1
        else:
            self._len = 0

        if params:
            self.__dict__.update(params)

    def __getitem__(self, idx):
        """Returns the members of the specified cluster.

        @param idx: the index of the cluster
        @return: the members of the specified cluster as a list
        @raise IndexError: if the index is out of bounds"""
        if idx < 0 or idx >= self._len:
            raise IndexError("cluster index out of range")
        return [i for i, e in enumerate(self._membership) if e == idx]

    def __iter__(self):
        """Iterates over the clusters in this clustering.

        This method will return a generator that generates the clusters
        one by one."""
        clusters = [[] for _ in range(self._len)]
        for idx, cluster in enumerate(self._membership):
            clusters[cluster].append(idx)
        return iter(clusters)

    def __len__(self):
        """Returns the number of clusters.

        @return: the number of clusters
        """
        return self._len

    def __str__(self):
        return self.summary(verbosity=1, width=78)

    def as_cover(self):
        """Returns a L{Cover} that contains the same clusters as this clustering."""
        return Cover(self._graph, self)

    def compare_to(self, other, *args, **kwds):
        """Compares this clustering to another one using some similarity or
        distance metric.

        This is a convenience method that simply calls L{compare_communities}
        with the two clusterings as arguments. Any extra positional or keyword
        argument is also forwarded to L{compare_communities}."""
        return compare_communities(self, other, *args, **kwds)

    @property
    def membership(self):
        """Returns the membership vector."""
        return self._membership[:]

    @property
    def n(self):
        """Returns the number of elements covered by this clustering."""
        return len(self._membership)

    def size(self, idx):
        """Returns the size of a given cluster.

        @param idx: the cluster in which we are interested.
        """
        return len(self[idx])

    def sizes(self, *args):
        """Returns the size of given clusters.

        The indices are given as positional arguments. If there are no
        positional arguments, the function will return the sizes of all clusters.
        """
        counts = [0] * len(self)
        for x in self._membership:
            if x is not None:
                counts[x] += 1

        if args:
            return [counts[idx] for idx in args]

        return counts

    def size_histogram(self, bin_width=1):
        """Returns the histogram of cluster sizes.

        @param bin_width: the bin width of the histogram
        @return: a L{Histogram} object
        """
        return Histogram(bin_width, self.sizes())

    def summary(self, verbosity=0, width=None):
        """Returns the summary of the clustering.

        The summary includes the number of items and clusters, and also the
        list of members for each of the clusters if the verbosity is nonzero.

        @param verbosity: determines whether the cluster members should be
          printed. Zero verbosity prints the number of items and clusters only.
        @return: the summary of the clustering as a string.
        """
        out = StringIO()
        print(
            "Clustering with %d elements and %d clusters"
            % (
                len(self._membership),
                len(self),
            ),
            file=out,
        )

        if verbosity < 1:
            return out.getvalue().strip()

        ndigits = len(str(len(self)))
        wrapper = _get_wrapper_for_width(width, subsequent_indent=" " * (ndigits + 3))

        for idx, cluster in enumerate(self._formatted_cluster_iterator()):
            wrapper.initial_indent = "[%*d] " % (ndigits, idx)
            print("\n".join(wrapper.wrap(cluster)), file=out)

        return out.getvalue().strip()

    def _formatted_cluster_iterator(self):
        """Iterates over the clusters and formats them into a string to be
        presented in the summary."""
        for cluster in self:
            yield ", ".join(str(member) for member in cluster)


class VertexClustering(Clustering):
    """The clustering of the vertex set of a graph.

    This class extends L{Clustering} by linking it to a specific L{Graph} object
    and by optionally storing the modularity score of the clustering.
    It also provides some handy methods like getting the subgraph corresponding
    to a cluster and such.

    @note: since this class is linked to a L{Graph}, destroying the graph by the
      C{del} operator does not free the memory occupied by the graph if there
      exists a L{VertexClustering} that references the L{Graph}.
    """

    # Allow None to be passed to __plot__ as the "palette" keyword argument
    _default_palette = None

    def __init__(
        self,
        graph,
        membership=None,
        modularity=None,
        params=None,
        modularity_params=None,
    ):
        """Creates a clustering object for a given graph.

        @param graph: the graph that will be associated to the clustering
        @param membership: the membership list. The length of the list must
          be equal to the number of vertices in the graph. If C{None}, every
          vertex is assumed to belong to the same cluster.
        @param modularity: the modularity score of the clustering. If C{None},
          it will be calculated when needed.
        @param params: additional parameters to be stored in this object.
        @param modularity_params: arguments that should be passed to
          L{Graph.modularity} when the modularity is (re)calculated. If the
          original graph was weighted, you should pass a dictionary
          containing a C{weight} key with the appropriate value here.
        """
        if membership is None:
            super().__init__([0] * graph.vcount(), params)
        else:
            if len(membership) != graph.vcount():
                raise ValueError("membership list has invalid length")
            super().__init__(membership, params)

        self._graph = graph
        self._modularity = modularity
        self._modularity_dirty = modularity is None
        if modularity_params is None:
            self._modularity_params = {}
        else:
            self._modularity_params = dict(modularity_params)

    @classmethod
    def FromAttribute(cls, graph, attribute, intervals=None, params=None):
        """Creates a vertex clustering based on the value of a vertex attribute.

        Vertices having the same attribute will correspond to the same cluster.

        @param graph: the graph on which we are working
        @param attribute: name of the attribute on which the clustering
            is based.
        @param intervals: for numeric attributes, you can either pass a single
            number or a list of numbers here. A single number means that the
            vertices will be put in bins of that width and vertices ending up
            in the same bin will be in the same cluster. A list of numbers
            specify the bin positions explicitly; e.g., C{[10, 20, 30]} means
            that there will be four categories: vertices with the attribute
            value less than 10, between 10 and 20, between 20 and 30 and over 30.
            Intervals are closed from the left and open from the right.
        @param params: additional parameters to be stored in this object.

        @return: a new VertexClustering object
        """
        from bisect import bisect

        def safeintdiv(x, y):
            """Safe integer division that handles None gracefully"""
            if x is None:
                return None
            return int(x / y)

        def safebisect(intervals, x):
            """Safe list bisection that handles None gracefully"""
            if x is None:
                return None
            return bisect(intervals, x)

        try:
            _ = iter(intervals)
            iterable = True
        except TypeError:
            iterable = False
        if intervals is None:
            vec = graph.vs[attribute]
        elif iterable:
            intervals = list(intervals)
            vec = [safebisect(intervals, x) for x in graph.vs[attribute]]
        else:
            intervals = float(intervals)
            vec = [safeintdiv(x, intervals) for x in graph.vs[attribute]]

        idgen = UniqueIdGenerator()
        idgen[None] = None
        vec = [idgen[i] for i in vec]
        return cls(graph, vec, None, params)

    def as_cover(self):
        """Returns a L{VertexCover} that contains the same clusters as this
        clustering."""
        return VertexCover(self._graph, self)

    def cluster_graph(self, combine_vertices=None, combine_edges=None):
        """Returns a graph where each cluster is contracted into a single
        vertex.

        In the resulting graph, vertex M{i} represents cluster M{i} in this
        clustering. Vertex M{i} and M{j} will be connected if there was
        at least one connected vertex pair M{(a, b)} in the original graph such
        that vertex M{a} was in cluster M{i} and vertex M{b} was in cluster
        M{j}.

        @param combine_vertices: specifies how to derive the attributes of
          the vertices in the new graph from the attributes of the old ones.
          See L{Graph.contract_vertices()<igraph._igraph.GraphBase.contract_vertices>}
          for more details.
        @param combine_edges: specifies how to derive the attributes of the
          edges in the new graph from the attributes of the old ones. See
          L{Graph.simplify()<igraph._igraph.GraphBase.simplify>} for more details.
          If you specify C{False} here, edges will not be combined, and the
          number of edges between the vertices representing the original
          clusters will be equal to the number of edges between the members of
          those clusters in the original graph.

        @return: the new graph.
        """
        result = self.graph.copy()
        result.contract_vertices(self.membership, combine_vertices)
        if combine_edges is not False:
            result.simplify(combine_edges=combine_edges)
        return result

    def crossing(self):
        """Returns a boolean vector where element M{i} is C{True} iff edge
        M{i} lies between clusters, C{False} otherwise."""
        membership = self.membership
        return [
            membership[v1] != membership[v2] for v1, v2 in self.graph.get_edgelist()
        ]

    @property
    def modularity(self):
        """Returns the modularity score"""
        if self._modularity_dirty:
            return self._recalculate_modularity_safe()
        return self._modularity

    q = modularity

    @property
    def graph(self):
        """Returns the graph belonging to this object"""
        return self._graph

    def recalculate_modularity(self):
        """Recalculates the stored modularity value.

        This method must be called before querying the modularity score of the
        clustering through the class member C{modularity} or C{q} if the
        graph has been modified (edges have been added or removed) since the
        creation of the L{VertexClustering} object.

        @return: the new modularity score
        """
        self._modularity = self._graph.modularity(
            self._membership, **self._modularity_params
        )
        self._modularity_dirty = False
        return self._modularity

    def _recalculate_modularity_safe(self):
        """Recalculates the stored modularity value and swallows all exceptions
        raised by the modularity function (if any).

        @return: the new modularity score or C{None} if the modularity function
        could not be calculated.
        """
        try:
            return self.recalculate_modularity()
        except Exception:
            return None
        finally:
            self._modularity_dirty = False

    def subgraph(self, idx):
        """Get the subgraph belonging to a given cluster.

        Precondition: the vertex set of the graph hasn't been modified since the
        moment the cover was constructed.

        @param idx: the cluster index
        @return: a copy of the subgraph
        """
        return self._graph.subgraph(self[idx])

    def subgraphs(self):
        """Gets all the subgraphs belonging to each of the clusters.

        Precondition: the vertex set of the graph hasn't been modified since the
        moment the cover was constructed.

        @return: a list containing copies of the subgraphs
        """
        return [self._graph.subgraph(cl) for cl in self]

    def giant(self):
        """Returns the largest cluster of the clustered graph.

        The largest cluster is a cluster for which no larger cluster exists in
        the clustering. It may also be known as the I{giant community} if the
        clustering represents the result of a community detection function.

        Precondition: the vertex set of the graph hasn't been modified since the
        moment the cover was constructed.

        @note: there can be multiple largest clusters, this method will return
          the copy of an arbitrary one if there are multiple largest clusters.

        @return: a copy of the largest cluster.
        """
        ss = self.sizes()
        if ss:
            max_size = max(ss)
            return self.subgraph(ss.index(max_size))
        else:
            return self._graph.copy()

    def __plot__(self, backend, context, *args, **kwds):
        """Plots the clustering to the given Cairo context or matplotlib Axes.

        This is done by calling L{Graph.__plot__()} with the same arguments, but
        coloring the graph vertices according to the current clustering (unless
        overridden by the C{vertex_color} argument explicitly).

        This method understands all the positional and keyword arguments that
        are understood by L{Graph.__plot__()}, only the differences will be
        highlighted here:

          - C{mark_groups}: whether to highlight some of the vertex groups by
            colored polygons. Besides the values accepted by L{Graph.__plot__}
            (i.e., a dict mapping colors to vertex indices, a list containing
            lists of vertex indices, or C{False}), the following are also
            accepted:

              - C{True}: all the groups will be highlighted, the colors matching
                the corresponding color indices from the current palette
                (see the C{palette} keyword argument of L{Graph.__plot__}).

              - A dict mapping cluster indices or tuples of vertex indices to
                color names.  The given clusters or vertex groups will be
                highlighted by the given colors.

              - A list of cluster indices. This is equivalent to passing a
                dict mapping numeric color indices from the current palette
                to cluster indices; therefore, the cluster referred to by element
                I{i} of the list will be highlighted by color I{i} from the
                palette.

            The value of the C{plotting.mark_groups} configuration key is also
            taken into account here; if that configuration key is C{True} and
            C{mark_groups} is not given explicitly, it will automatically be set
            to C{True}.

            In place of lists of vertex indices, you may also use L{VertexSeq}
            instances.

            In place of color names, you may also use color indices into the
            current palette. C{None} as a color name will mean that the
            corresponding group is ignored.

          - C{palette}: the palette used to resolve numeric color indices to RGBA
            values. By default, this is an instance of L{ClusterColoringPalette}.

        @see: L{Graph.__plot__()} for more supported keyword arguments.
        """
        from igraph.drawing.colors import default_edge_colors

        if "edge_color" not in kwds and "color" not in self.graph.edge_attributes():
            # Set up a default edge coloring based on internal vs external edges
            colors = default_edge_colors[backend]
            kwds["edge_color"] = [
                colors[is_crossing] for is_crossing in self.crossing()
            ]

        palette = kwds.get("palette", None)
        if palette is None:
            kwds["palette"] = ClusterColoringPalette(len(self))

        if "mark_groups" not in kwds:
            if Configuration.instance()["plotting.mark_groups"]:
                kwds["mark_groups"] = self
        else:
            kwds["mark_groups"] = _handle_mark_groups_arg_for_clustering(
                kwds["mark_groups"], self
            )

        if "vertex_color" not in kwds:
            kwds["vertex_color"] = self.membership
        result = self._graph.__plot__(backend, context, *args, **kwds)

        return result

    def _formatted_cluster_iterator(self):
        """Iterates over the clusters and formats them into a string to be
        presented in the summary."""
        if self._graph.is_named():
            names = self._graph.vs["name"]
            for cluster in self:
                yield ", ".join(str(names[member]) for member in cluster)
        else:
            for cluster in self:
                yield ", ".join(str(member) for member in cluster)


###############################################################################


class Dendrogram:
    """The hierarchical clustering (dendrogram) of some dataset.

    A hierarchical clustering means that we know not only the way the
    elements are separated into groups, but also the exact history of
    how individual elements were joined into larger subgroups.

    This class internally represents the hierarchy by a matrix with n rows
    and 2 columns -- or more precisely, a list of lists of size 2. This is
    exactly the same as the original format used by C{igraph}'s C core.
    The M{i}th row of the matrix contains the indices of the two clusters
    being joined in time step M{i}. The joint group will be represented by
    the ID M{n+i}, with M{i} starting from one. The ID of the joint group
    will be referenced in the upcoming steps instead of any of its individual
    members. So, IDs less than or equal to M{n} (where M{n} is the number of
    rows in the matrix) mean the original members of the dataset (with ID
    from 0 to M{n}), while IDs up from M{n+1} mean joint groups. As an
    example, take a look at the dendrogram and the internal representation of
    a given clustering of five nodes::

      0 -+
         |
      1 -+-+
           |
      2 ---+-+        <====>   [[0, 1], [3, 4], [2, 5], [6, 7]]
             |
      3 -+   |
         |   |
      4 -+---+---
    """

    def __init__(self, merges):
        """Creates a hierarchical clustering.

        @param merges: the merge history either in matrix or tuple format"""
        self._merges = [tuple(pair) for pair in merges]
        self._nmerges = len(self._merges)
        if self._nmerges:
            self._nitems = max(self._merges[-1]) - self._nmerges + 2
        else:
            self._nitems = 0
        self._names = None

    @staticmethod
    def _convert_matrix_to_tuple_repr(merges, n=None):
        """Converts the matrix representation of a clustering to a tuple
        representation.

        @param merges: the matrix representation of the clustering
        @return: the tuple representation of the clustering
        """
        if n is None:
            n = len(merges) + 1
        tuple_repr = range(n)
        idxs = list(range(n))
        for rowidx, row in enumerate(merges):
            i, j = row
            try:
                idxi, idxj = idxs[i], idxs[j]
                tuple_repr[idxi] = (tuple_repr[idxi], tuple_repr[idxj])
                tuple_repr[idxj] = None
            except IndexError:
                raise ValueError(
                    "malformed matrix, subgroup referenced "
                    + "before being created in step %d" % rowidx
                ) from None
            idxs.append(j)
        return [x for x in tuple_repr if x is not None]

    def _traverse_inorder(self):
        """Conducts an inorder traversal of the merge tree.

        The inorder traversal returns the nodes on the last level in the order
        they should be drawn so that no edges cross each other.

        @return: the result of the inorder traversal in a list."""
        result = []
        seen_nodes = set()

        for node_index in reversed(range(self._nitems + self._nmerges)):
            if node_index in seen_nodes:
                continue

            stack = [node_index]
            while stack:
                last = stack.pop()
                seen_nodes.add(last)
                if last < self._nitems:
                    # 'last' is a regular node so the traversal ends here, we
                    # can append it to the results
                    result.append(last)
                else:
                    # 'last' is a merge node, so let us proceed with the entry
                    # where this merge node was created
                    stack.extend(self._merges[last - self._nitems])

        return result

    def __str__(self):
        return self.summary(verbosity=1)

    def format(self, format="newick"):
        """Formats the dendrogram in a foreign format.

        Currently only the Newick format is supported.

        Example:

            >>> d = Dendrogram([(2, 3), (0, 1), (4, 5)])
            >>> d.format()
            '((2,3)4,(0,1)5)6;'
            >>> d.names = list("ABCDEFG")
            >>> d.format()
            '((C,D)E,(A,B)F)G;'
        """
        if format == "newick":
            n = self._nitems + self._nmerges
            if self._names is None:
                nodes = list(range(n))
            else:
                nodes = list(self._names)
            if len(nodes) < n:
                nodes.extend("" for _ in range(n - len(nodes)))
            for k, (i, j) in enumerate(self._merges, self._nitems):
                nodes[k] = "(%s,%s)%s" % (nodes[i], nodes[j], nodes[k])
                nodes[i] = nodes[j] = None
            return nodes[-1] + ";"
        raise ValueError("unsupported format: %r" % format)

    def summary(self, verbosity=0, max_leaf_count=40):
        """Returns the summary of the dendrogram.

        The summary includes the number of leafs and branches, and also an
        ASCII art representation of the dendrogram unless it is too large.

        @param verbosity: determines whether the ASCII representation of the
          dendrogram should be printed. Zero verbosity prints only the number
          of leafs and branches.
        @param max_leaf_count: the maximal number of leafs to print in the
          ASCII representation. If the dendrogram has more leafs than this
          limit, the ASCII representation will not be printed even if the
          verbosity is larger than or equal to 1.
        @return: the summary of the dendrogram as a string.
        """
        out = StringIO()
        print(
            "Dendrogram, %d elements, %d merges"
            % (
                self._nitems,
                self._nmerges,
            ),
            file=out,
        )

        if self._nitems == 0 or verbosity < 1 or self._nitems > max_leaf_count:
            return out.getvalue().strip()

        print("", file=out)

        positions = [None] * self._nitems
        inorder = self._traverse_inorder()
        distance = 2
        level_distance = 2
        nextp = 0
        for idx, element in enumerate(inorder):
            positions[element] = nextp
            inorder[idx] = str(element)
            nextp += max(distance, len(inorder[idx]) + 1)

        width = max(positions) + 1

        # Print the nodes on the lowest level
        print((" " * (distance - 1)).join(inorder), file=out)
        midx = 0
        max_community_idx = self._nitems
        while midx < self._nmerges:
            char_array = [" "] * width
            for position in positions:
                if position >= 0:
                    char_array[position] = "|"
            char_str = "".join(char_array)
            for _ in range(level_distance - 1):
                print(char_str, file=out)  # Print the lines

            cidx_incr = 0
            while midx < self._nmerges:
                id1, id2 = self._merges[midx]
                if id1 >= max_community_idx or id2 >= max_community_idx:
                    break
                midx += 1

                pos1, pos2 = positions[id1], positions[id2]
                positions[id1], positions[id2] = -1, -1

                if pos1 > pos2:
                    pos1, pos2 = pos2, pos1
                positions.append((pos1 + pos2) // 2)

                dashes = "-" * (pos2 - pos1 - 1)
                char_array[pos1 : (pos2 + 1)] = "`%s'" % dashes

                cidx_incr += 1

            max_community_idx += cidx_incr

            print("".join(char_array), file=out)

        return out.getvalue().strip()

    def __plot__(self, backend, context, *args, **kwds):
        """Draws the dendrogram on the given Cairo context or matplotlib Axes.

        Supported keyword arguments are:

          - C{orientation}: the orientation of the dendrogram. Must be one of
            the following values: C{left-right}, C{bottom-top}, C{right-left}
            or C{top-bottom}. Individual elements are always placed at the
            former edge and merges are performed towards the latter edge.
            Possible aliases: C{horizontal} = C{left-right},
            C{vertical} = C{bottom-top}, C{lr} = C{left-right},
            C{rl} = C{right-left}, C{tb} = C{top-bottom}, C{bt} = C{bottom-top}.
            The default is C{left-right}.

        """
        if backend == "matplotlib":
            drawer = MatplotlibDendrogramDrawer(context)
        else:
            bbox = kwds.pop("bbox", None)
            palette = kwds.pop("palette", None)
            if bbox is None:
                raise ValueError("bbox is required for Cairo plots")
            if palette is None:
                raise ValueError("palette is required for Cairo plots")
            drawer = CairoDendrogramDrawer(context, bbox, palette)

        drawer.draw(self, **kwds)

    @property
    def merges(self):
        """Returns the performed merges in matrix format"""
        return deepcopy(self._merges)

    @property
    def names(self):
        """Returns the names of the nodes in the dendrogram"""
        return self._names

    @names.setter
    def names(self, items):
        """Sets the names of the nodes in the dendrogram"""
        if items is None:
            self._names = None
            return

        items = list(items)
        if len(items) < self._nitems:
            raise ValueError("must specify at least %d names" % self._nitems)

        n = self._nitems + self._nmerges
        self._names = items[:n]
        if len(self._names) < n:
            self._names.extend("" for _ in range(n - len(self._names)))


class VertexDendrogram(Dendrogram):
    """The dendrogram resulting from the hierarchical clustering of the
    vertex set of a graph."""

    def __init__(self, graph, merges, optimal_count=None, modularity_params=None):
        """Creates a dendrogram object for a given graph.

        @param graph: the graph that will be associated to the clustering
        @param merges: the merges performed given in matrix form.
        @param optimal_count: the optimal number of clusters where the
          dendrogram should be cut. This is a hint usually provided by the
          clustering algorithm that produces the dendrogram. C{None} means
          that such a hint is not available; the optimal count will then be
          selected based on the modularity in such a case.
        @param modularity_params: arguments that should be passed to
          L{Graph.modularity} when the modularity is (re)calculated. If the
          original graph was weighted, you should pass a dictionary
          containing a C{weight} key with the appropriate value here.
        """
        super().__init__(merges)
        self._graph = graph
        self._optimal_count = optimal_count
        if modularity_params is None:
            self._modularity_params = {}
        else:
            self._modularity_params = dict(modularity_params)

    def as_clustering(self, n=None):
        """Cuts the dendrogram at the given level and returns a corresponding
        L{VertexClustering} object.

        @param n: the desired number of clusters. Merges are replayed from the
          beginning until the membership vector has exactly M{n} distinct elements
          or until there are no more recorded merges, whichever happens first.
          If C{None}, the optimal count hint given by the clustering algorithm
          will be used If the optimal count was not given either, it will be
          calculated by selecting the level where the modularity is maximal.
        @return: a new L{VertexClustering} object.
        """
        if n is None:
            n = self.optimal_count
        num_elts = self._graph.vcount()
        idgen = UniqueIdGenerator()
        membership = community_to_membership(self._merges, num_elts, num_elts - n)
        membership = [idgen[m] for m in membership]
        return VertexClustering(
            self._graph, membership, modularity_params=self._modularity_params
        )

    @property
    def optimal_count(self):
        """Returns the optimal number of clusters for this dendrogram.

        If an optimal count hint was given at construction time, this
        property simply returns the hint. If such a count was not given,
        this method calculates the optimal number of clusters by maximizing
        the modularity along all the possible cuts in the dendrogram.
        """
        if self._optimal_count is not None:
            return self._optimal_count

        n = self._graph.vcount()
        if n == 0:
            return 0

        max_q, optimal_count = 0, n - len(self._merges)
        for step in range(min(n - 1, len(self._merges) + 1)):
            membs = community_to_membership(self._merges, n, step)
            q = self._graph.modularity(membs, **self._modularity_params)
            if q > max_q:
                optimal_count = n - step
                max_q = q

        self._optimal_count = optimal_count
        return optimal_count

    @optimal_count.setter
    def optimal_count(self, value):
        self._optimal_count = max(int(value), 1)

    def __plot__(self, backend, context, *args, **kwds):
        """Draws the vertex dendrogram on the given Cairo context or matplotlib Axes

        See L{Dendrogram.__plot__} for the list of supported keyword
        arguments."""
        from igraph.drawing.metamagic import AttributeCollectorBase

        class VisualVertexBuilder(AttributeCollectorBase):
            _kwds_prefix = "vertex_"
            label = None

        builder = VisualVertexBuilder(self._graph.vs, kwds)
        self._names = [vertex.label for vertex in builder]
        self._names = [
            name if name is not None else str(idx)
            for idx, name in enumerate(self._names)
        ]
        result = Dendrogram.__plot__(self, backend, context, *args, **kwds)
        del self._names

        return result


###############################################################################


class Cover:
    """Class representing a cover of an arbitrary ordered set.

    Covers are similar to clusterings, but each element of the set may
    belong to more than one cluster in a cover, and elements not belonging
    to any cluster are also allowed.

    L{Cover} instances provide a similar API as L{Clustering} instances;
    for instance, iterating over a L{Cover} will iterate over the clusters
    just like with a regular L{Clustering} instance. However, they are not
    derived from each other or from a common superclass, and there might
    be functions that exist only in one of them or the other.

    Clusters of an individual cover can be accessed by the C{[]} operator:

      >>> cl = Cover([[0,1,2,3], [2,3,4], [0,1,6]])
      >>> cl[0]
      [0, 1, 2, 3]

    The membership vector can be accessed by the C{membership} property.
    Note that contrary to L{Clustering} instances, the membership vector
    will contain lists that contain the cluster indices each item belongs
    to:

      >>> cl.membership
      [[0, 2], [0, 2], [0, 1], [0, 1], [1], [], [2]]

    The number of clusters can be retrieved by the C{len} function:

      >>> len(cl)
      3

    You can iterate over the cover as if it were a regular list of
    clusters:

      >>> for cluster in cl:
      ...     print(" ".join(str(idx) for idx in cluster))
      ...
      0 1 2 3
      2 3 4
      0 1 6

    If you need all the clusters at once as lists, you can simply convert
    the cover to a list:

      >>> cluster_list = list(cl)
      >>> print(cluster_list)
      [[0, 1, 2, 3], [2, 3, 4], [0, 1, 6]]

    L{Clustering} objects can readily be converted to L{Cover} objects
    using the constructor:

      >>> clustering = Clustering([0, 0, 0, 0, 1, 1, 1, 2, 2, 2])
      >>> cover = Cover(clustering)
      >>> list(clustering) == list(cover)
      True
    """

    def __init__(self, clusters, n=0):
        """Constructs a cover with the given clusters.

        @param clusters: the clusters in this cover, as a list or iterable.
          Each cluster is specified by a list or tuple that contains the
          IDs of the items in this cluster. IDs start from zero.

        @param n: the total number of elements in the set that is covered
          by this cover. If it is less than the number of unique elements
          found in all the clusters, we will simply use the number of unique
          elements, so it is safe to leave this at zero. You only have to
          specify this parameter if there are some elements that are covered
          by none of the clusters.
        """

        self._clusters = [list(cluster) for cluster in clusters]
        try:
            self._n = max(max(cluster) + 1 for cluster in self._clusters if cluster)
        except ValueError:
            self._n = 0
        self._n = max(n, self._n)

    def __getitem__(self, index):
        """Returns the cluster with the given index."""
        return self._clusters[index]

    def __iter__(self):
        """Iterates over the clusters in this cover."""
        return iter(self._clusters)

    def __len__(self):
        """Returns the number of clusters in this cover."""
        return len(self._clusters)

    def __str__(self):
        """Returns a string representation of the cover."""
        return self.summary(verbosity=1, width=78)

    @property
    def membership(self):
        """Returns the membership vector of this cover.

        The membership vector of a cover covering I{n} elements is a list of
        length I{n}, where element I{i} contains the cluster indices of the
        I{i}th item.
        """
        result = [[] for _ in range(self._n)]
        for idx, cluster in enumerate(self):
            for item in cluster:
                result[item].append(idx)
        return result

    @property
    def n(self):
        """Returns the number of elements in the set covered by this cover."""
        return self._n

    def size(self, idx):
        """Returns the size of a given cluster.

        @param idx: the cluster in which we are interested.
        """
        return len(self[idx])

    def sizes(self, *args):
        """Returns the size of given clusters.

        The indices are given as positional arguments. If there are no
        positional arguments, the function will return the sizes of all clusters.
        """
        if args:
            return [len(self._clusters[idx]) for idx in args]
        return [len(cluster) for cluster in self]

    def size_histogram(self, bin_width=1):
        """Returns the histogram of cluster sizes.

        @param bin_width: the bin width of the histogram
        @return: a L{Histogram} object
        """
        return Histogram(bin_width, self.sizes())

    def summary(self, verbosity=0, width=None):
        """Returns the summary of the cover.

        The summary includes the number of items and clusters, and also the
        list of members for each of the clusters if the verbosity is nonzero.

        @param verbosity: determines whether the cluster members should be
          printed. Zero verbosity prints the number of items and clusters only.
        @return: the summary of the cover as a string.
        """
        out = StringIO()
        print("Cover with %d clusters" % len(self), file=out)

        if verbosity < 1:
            return out.getvalue().strip()

        ndigits = len(str(len(self)))
        wrapper = _get_wrapper_for_width(width, subsequent_indent=" " * (ndigits + 3))

        for idx, cluster in enumerate(self._formatted_cluster_iterator()):
            wrapper.initial_indent = "[%*d] " % (ndigits, idx)
            print("\n".join(wrapper.wrap(cluster)), file=out)

        return out.getvalue().strip()

    def _formatted_cluster_iterator(self):
        """Iterates over the clusters and formats them into a string to be
        presented in the summary."""
        for cluster in self:
            yield ", ".join(str(member) for member in cluster)


class VertexCover(Cover):
    """The cover of the vertex set of a graph.

    This class extends L{Cover} by linking it to a specific L{Graph} object.
    It also provides some handy methods like getting the subgraph corresponding
    to a cluster and such.

    @note: since this class is linked to a L{Graph}, destroying the graph by the
      C{del} operator does not free the memory occupied by the graph if there
      exists a L{VertexCover} that references the L{Graph}.
    """

    def __init__(self, graph, clusters=None):
        """Creates a cover object for a given graph.

        @param graph: the graph that will be associated to the cover
        @param clusters: the list of clusters. If C{None}, it is assumed
          that there is only a single cluster that covers the whole graph.
        """
        if clusters is None:
            clusters = [range(graph.vcount())]

        self._resolve_names_in_clusters(graph, clusters)

        super().__init__(clusters, n=graph.vcount())
        if self._n > graph.vcount():
            raise ValueError(
                "cluster list contains vertex ID larger than the "
                "number of vertices in the graph"
            )

        self._graph = graph

    def crossing(self):
        """Returns a boolean vector where element M{i} is C{True} iff edge
        M{i} lies between clusters, C{False} otherwise."""
        membership = [frozenset(cluster) for cluster in self.membership]
        return [
            membership[v1].isdisjoint(membership[v2])
            for v1, v2 in self.graph.get_edgelist()
        ]

    @property
    def graph(self):
        """Returns the graph belonging to this object"""
        return self._graph

    def subgraph(self, idx):
        """Get the subgraph belonging to a given cluster.

        Precondition: the vertex set of the graph hasn't been modified since the
        moment the cover was constructed.

        @param idx: the cluster index
        @return: a copy of the subgraph
        """
        return self._graph.subgraph(self[idx])

    def subgraphs(self):
        """Gets all the subgraphs belonging to each of the clusters.

        Precondition: the vertex set of the graph hasn't been modified since the
        moment the cover was constructed.

        @return: a list containing copies of the subgraphs
        """
        return [self._graph.subgraph(cl) for cl in self]

    def __plot__(self, backend, context, *args, **kwds):
        """Plots the cover to the given Cairo context or matplotlib Axes.

        This is done by calling L{Graph.__plot__()} with the same arguments, but
        drawing nice colored blobs around the vertex groups.

        This method understands all the positional and keyword arguments that
        are understood by L{Graph.__plot__()}, only the differences will be
        highlighted here:

          - C{mark_groups}: whether to highlight the vertex clusters by
            colored polygons. Besides the values accepted by L{Graph.__plot__}
            (i.e., a dict mapping colors to vertex indices, a list containing
            lists of vertex indices, or C{False}), the following are also
            accepted:

              - C{True}: all the clusters will be highlighted, the colors matching
                the corresponding color indices from the current palette
                (see the C{palette} keyword argument of L{Graph.__plot__}).

              - A dict mapping cluster indices or tuples of vertex indices to
                color names.  The given clusters or vertex groups will be
                highlighted by the given colors.

              - A list of cluster indices. This is equivalent to passing a
                dict mapping numeric color indices from the current palette
                to cluster indices; therefore, the cluster referred to by element
                I{i} of the list will be highlighted by color I{i} from the
                palette.

            The value of the C{plotting.mark_groups} configuration key is also
            taken into account here; if that configuration key is C{True} and
            C{mark_groups} is not given explicitly, it will automatically be set
            to C{True}.

            In place of lists of vertex indices, you may also use L{VertexSeq}
            instances.

            In place of color names, you may also use color indices into the
            current palette. C{None} as a color name will mean that the
            corresponding group is ignored.

          - C{palette}: the palette used to resolve numeric color indices to RGBA
            values. By default, this is an instance of L{ClusterColoringPalette}.

        @see: L{Graph.__plot__()} for more supported keyword arguments.
        """
        if "edge_color" not in kwds and "color" not in self.graph.edge_attributes():
            # Set up a default edge coloring based on internal vs external edges
            if backend == "matplotlib":
                colors = ["dimgrey", "silver"]
            else:
                colors = ["grey20", "grey80"]

            kwds["edge_color"] = [
                colors[is_crossing] for is_crossing in self.crossing()
            ]

        palette = kwds.get("palette", None)
        if palette is None:
            kwds["palette"] = ClusterColoringPalette(len(self))

        if "mark_groups" not in kwds:
            if Configuration.instance()["plotting.mark_groups"]:
                kwds["mark_groups"] = self
        else:
            kwds["mark_groups"] = _handle_mark_groups_arg_for_clustering(
                kwds["mark_groups"], self
            )

        return self._graph.__plot__(backend, context, *args, **kwds)

    def _formatted_cluster_iterator(self):
        """Iterates over the clusters and formats them into a string to be
        presented in the summary."""
        if self._graph.is_named():
            names = self._graph.vs["name"]
            for cluster in self:
                yield ", ".join(str(names[member]) for member in cluster)
        else:
            for cluster in self:
                yield ", ".join(str(member) for member in cluster)

    @staticmethod
    def _resolve_names_in_clusters(graph, clusters):
        if not graph.is_named():
            return

        names = graph.vs["name"]
        name_to_index = {k: v for v, k in enumerate(names)}

        for idx, cluster in enumerate(clusters):
            if any(isinstance(item, str) for item in cluster):
                new_cluster = []
                for item in cluster:
                    if isinstance(item, str):
                        new_cluster.append(name_to_index.get(item, item))
                    else:
                        new_cluster.append(item)
                clusters[idx] = new_cluster


class CohesiveBlocks(VertexCover):
    """The cohesive block structure of a graph.

    Instances of this type are created by
    L{Graph.cohesive_blocks()<GraphBase.cohesive_blocks()>}. See the
    documentation of L{Graph.cohesive_blocks()<GraphBase.cohesive_blocks()>}
    for an explanation of what cohesive blocks are.

    This class provides a few more methods that make handling of cohesive
    block structures easier.
    """

    def __init__(self, graph, blocks=None, cohesion=None, parent=None):
        """Constructs a new cohesive block structure for the given graph.

        If any of I{blocks}, I{cohesion} or I{parent} is C{None}, all the
        arguments will be ignored and
        L{Graph.cohesive_blocks()<GraphBase.cohesive_blocks()>} will be called
        to calculate the cohesive blocks. Otherwise, these three variables
        should describe the *result* of a cohesive block structure calculation.
        Chances are that you never have to construct L{CohesiveBlocks}
        instances directly, just use
        L{Graph.cohesive_blocks()<GraphBase.cohesive_blocks()>}.

        @param graph: the graph itself
        @param blocks: a list containing the blocks; each block is described
          as a list containing vertex IDs.
        @param cohesion: the cohesion of each block. The length of this list
          must be equal to the length of I{blocks}.
        @param parent: the parent block of each block. Negative values or
          C{None} mean that there is no parent block for that block. There
          should be only one parent block, which covers the entire graph.
        @see: Graph.cohesive_blocks()
        """
        if blocks is None or cohesion is None or parent is None:
            blocks, cohesion, parent = graph.cohesive_blocks()

        super().__init__(graph, blocks)

        self._cohesion = cohesion
        self._parent = parent
        for idx, p in enumerate(self._parent):
            if p < 0:
                self._parent[idx] = None

    def cohesion(self, idx):
        """Returns the cohesion of the group with the given index."""
        return self._cohesion[idx]

    def cohesions(self):
        """Returns the list of cohesion values for each group."""
        return self._cohesion[:]

    def hierarchy(self):
        """Returns a new graph that describes the hierarchical relationships
        between the groups.

        The new graph will be a directed tree; an edge will point from
        vertex M{i} to vertex M{j} if group M{i} is a superset of group M{j}.
        In other words, the edges point downwards.
        """
        from igraph import Graph

        edges = [
            pair for pair in zip(self._parent, range(len(self))) if pair[0] is not None
        ]
        return Graph(edges, directed=True)

    def max_cohesion(self, idx):
        """Finds the maximum cohesion score among all the groups that contain
        the given vertex."""
        result = 0
        for cohesion, cluster in zip(self._cohesion, self._clusters):
            if idx in cluster:
                result = max(result, cohesion)
        return result

    def max_cohesions(self):
        """For each vertex in the graph, returns the maximum cohesion score
        among all the groups that contain the vertex."""
        result = [0] * self._graph.vcount()
        for cohesion, cluster in zip(self._cohesion, self._clusters):
            for idx in cluster:
                result[idx] = max(result[idx], cohesion)
        return result

    def parent(self, idx):
        """Returns the parent group index of the group with the given index
        or C{None} if the given group is the root."""
        return self._parent[idx]

    def parents(self):
        """Returns the list of parent group indices for each group or C{None}
        if the given group is the root."""
        return self._parent[:]

    def __plot__(self, backend, context, *args, **kwds):
        """Plots the cohesive block structure to the given Cairo context or
        matplotlib Axes.

        Since a L{CohesiveBlocks} instance is also a L{VertexCover}, keyword
        arguments accepted by L{VertexCover.__plot__()} are also accepted here.
        The only difference is that the vertices are colored according to their
        maximal cohesions by default, and groups are marked by colored blobs
        except the last group which encapsulates the whole graph.

        See the documentation of L{VertexCover.__plot__()} for more details.
        """
        prepare_groups = False
        if "mark_groups" not in kwds:
            if Configuration.instance()["plotting.mark_groups"]:
                prepare_groups = True
        elif kwds["mark_groups"] is True:
            prepare_groups = True

        if prepare_groups:
            colors = [pair for pair in enumerate(self.cohesions()) if pair[1] > 1]
            kwds["mark_groups"] = colors

        if "vertex_color" not in kwds:
            kwds["vertex_color"] = self.max_cohesions()

        return VertexCover.__plot__(self, backend, context, *args, **kwds)


def _handle_mark_groups_arg_for_clustering(mark_groups, clustering):
    """Handles the mark_groups=... keyword argument in plotting methods of
    clusterings.

    This is an internal method, you shouldn't need to mess around with it.
    Its purpose is to handle the extended semantics of the mark_groups=...
    keyword argument in the C{__plot__} method of L{VertexClustering} and
    L{VertexCover} instances, namely the feature that numeric IDs are resolved
    to clusters automatically.
    """
    # Handle the case of mark_groups = True, mark_groups containing a list or
    # tuple of cluster IDs, and mark_groups yielding (cluster ID, color)
    # pairs
    if mark_groups is True:
        group_iter = ((group, color) for color, group in enumerate(clustering))
    elif isinstance(mark_groups, dict):
        group_iter = mark_groups.items()
    elif hasattr(mark_groups, "__getitem__") and hasattr(mark_groups, "__len__"):
        # Lists, tuples
        try:
            first = mark_groups[0]
        except Exception:
            # Hmm. Maybe not a list or tuple?
            first = None
        if first is not None:
            # Okay. Is the first element of the list a single number?
            if isinstance(first, int):
                # Yes. Seems like we have a list of cluster indices.
                # Assign color indices automatically.
                group_iter = ((group, color) for color, group in enumerate(mark_groups))
            else:
                # No. Seems like we have good ol' group-color pairs.
                group_iter = mark_groups
        else:
            group_iter = mark_groups
    elif hasattr(mark_groups, "__iter__"):
        # Iterators etc
        group_iter = mark_groups
    else:
        group_iter = {}.items()

    def cluster_index_resolver():
        for group, color in group_iter:
            if isinstance(group, int):
                group = clustering[group]
            yield group, color

    return cluster_index_resolver()


##############################################################


def _prepare_community_comparison(comm1, comm2, remove_none=False):
    """Auxiliary method that takes two community structures either as
    membership lists or instances of L{Clustering}, and returns a
    tuple whose two elements are membership lists.

    This is used by L{compare_communities} and L{split_join_distance}.

    @param comm1: the first community structure as a membership list or
      as a L{Clustering} object.
    @param comm2: the second community structure as a membership list or
      as a L{Clustering} object.
    @param remove_none: whether to remove C{None} entries from the membership
      lists. If C{remove_none} is C{False}, a C{None} entry in either C{comm1}
      or C{comm2} will result in an exception. If C{remove_none} is C{True},
      C{None} values are filtered away and only the remaining lists are
      compared.
    """

    def _ensure_list(obj):
        if isinstance(obj, Clustering):
            return obj.membership
        return list(obj)

    vec1, vec2 = _ensure_list(comm1), _ensure_list(comm2)
    if len(vec1) != len(vec2):
        raise ValueError("the two membership vectors must be equal in length")

    if remove_none and (None in vec1 or None in vec2):
        idxs_to_remove = [
            i for i in range(len(vec1)) if vec1[i] is None or vec2[i] is None
        ]
        idxs_to_remove.reverse()
        n = len(vec1)
        for i in idxs_to_remove:
            n -= 1
            vec1[i], vec1[n] = vec1[n], vec1[i]
            vec2[i], vec2[n] = vec2[n], vec2[i]
        del vec1[n:]
        del vec2[n:]

    return vec1, vec2


def compare_communities(comm1, comm2, method="vi", remove_none=False):
    """Compares two community structures using various distance measures.

    For measures involving entropies (e.g., the variation of information metric),
    igraph uses natural logarithms.

    B{References}

      - Meila M: Comparing clusterings by the variation of information. In:
        Scholkopf B, Warmuth MK (eds). Learning Theory and Kernel Machines: 16th
        Annual Conference on Computational Learning Theory and 7th Kernel
        Workship, COLT/Kernel 2003, Washington, DC, USA. Lecture Notes in Computer
        Science, vol. 2777, Springer, 2003. ISBN: 978-3-540-40720-1.
      - Danon L, Diaz-Guilera A, Duch J, Arenas A: Comparing community structure
        identification. I{J Stat Mech} P09008, 2005.
      - van Dongen S: Performance criteria for graph clustering and Markov
        cluster experiments. Technical Report INS-R0012, National Research
        Institute for Mathematics and Computer Science in the Netherlands,
        Amsterdam, May 2000.
      - Rand WM: Objective criteria for the evaluation of clustering
        methods. I{J Am Stat Assoc} 66(336):846-850, 1971.
      - Hubert L and Arabie P: Comparing partitions. I{Journal of
        Classification} 2:193-218, 1985.

    @param comm1: the first community structure as a membership list or
      as a L{Clustering} object.
    @param comm2: the second community structure as a membership list or
      as a L{Clustering} object.
    @param method: the measure to use. C{"vi"} or C{"meila"} means the
      variation of information metric of Meila (2003), C{"nmi"} or C{"danon"}
      means the normalized mutual information as defined by Danon et al. (2005),
      C{"split-join"} means the split-join distance of van Dongen (2000),
      C{"rand"} means the Rand index of Rand (1971), C{"adjusted_rand"}
      means the adjusted Rand index of Hubert and Arabie (1985).
    @param remove_none: whether to remove C{None} entries from the membership
      lists. This is handy if your L{Clustering} object was constructed using
      L{VertexClustering.FromAttribute} using an attribute which was not defined
      for all the vertices. If C{remove_none} is C{False}, a C{None} entry in
      either C{comm1} or C{comm2} will result in an exception. If C{remove_none}
      is C{True}, C{None} values are filtered away and only the remaining lists
      are compared.

    @return: the calculated measure.
    """
    vec1, vec2 = _prepare_community_comparison(comm1, comm2, remove_none)
    return _compare_communities(vec1, vec2, method)


def split_join_distance(comm1, comm2, remove_none=False):
    """Calculates the split-join distance between two community structures.

    The split-join distance is a distance measure defined on the space of
    partitions of a given set. It is the sum of the projection distance of
    one partition from the other and vice versa, where the projection
    number of A from B is if calculated as follows:

      1. For each set in A, find the set in B with which it has the
         maximal overlap, and take note of the size of the overlap.

      2. Take the sum of the maximal overlap sizes for each set in A.

      3. Subtract the sum from M{n}, the number of elements in the
         partition.

    Note that the projection distance is asymmetric, that's why it has to be
    calculated in both directions and then added together.  This function
    returns the projection distance of C{comm1} from C{comm2} and the
    projection distance of C{comm2} from C{comm1}, and returns them in a pair.
    The actual split-join distance is the sum of the two distances. The reason
    why it is presented this way is that one of the elements being zero then
    implies that one of the partitions is a subpartition of the other (and if
    it is close to zero, then one of the partitions is close to being a
    subpartition of the other).

    B{Reference}: van Dongen S: Performance criteria for graph clustering and
    Markov cluster experiments. Technical Report INS-R0012, National Research
    Institute for Mathematics and Computer Science in the Netherlands,
    Amsterdam, May 2000.

    @param comm1: the first community structure as a membership list or
      as a L{Clustering} object.
    @param comm2: the second community structure as a membership list or
      as a L{Clustering} object.
    @param remove_none: whether to remove C{None} entries from the membership
      lists. This is handy if your L{Clustering} object was constructed using
      L{VertexClustering.FromAttribute} using an attribute which was not defined
      for all the vertices. If C{remove_none} is C{False}, a C{None} entry in
      either C{comm1} or C{comm2} will result in an exception. If C{remove_none}
      is C{True}, C{None} values are filtered away and only the remaining lists
      are compared.

    @return: the projection distance of C{comm1} from C{comm2} and vice versa
      in a tuple. The split-join distance is the sum of the two.

    @see: L{compare_communities()} with C{method = "split-join"} if you are
      not interested in the individual projection distances but only the
      sum of them.
    """
    import igraph._igraph

    vec1, vec2 = _prepare_community_comparison(comm1, comm2, remove_none)
    return igraph._igraph._split_join_distance(vec1, vec2)


def _biconnected_components(graph, return_articulation_points=False):
    """\
    Calculates the biconnected components of the graph.

    @param return_articulation_points: whether to return the articulation
      points as well
    @return: a L{VertexCover} object describing the biconnected components,
      and optionally the list of articulation points as well
    """
    if return_articulation_points:
        trees, aps = GraphBase.biconnected_components(graph, True)
    else:
        trees = GraphBase.biconnected_components(graph, False)

    clusters = []
    if trees:
        edgelist = graph.get_edgelist()
        for tree in trees:
            cluster = set()
            for edge_id in tree:
                cluster.update(edgelist[edge_id])
            clusters.append(sorted(cluster))

    clustering = VertexCover(graph, clusters)

    if return_articulation_points:
        return clustering, aps
    else:
        return clustering


def _cohesive_blocks(graph):
    """Calculates the cohesive block structure of the graph.

    Cohesive blocking is a method of determining hierarchical subsets of graph
    vertices based on their structural cohesion (i.e. vertex connectivity).
    For a given graph G, a subset of its vertices S is said to be maximally
    k-cohesive if there is no superset of S with vertex connectivity greater
    than or equal to k. Cohesive blocking is a process through which, given a
    k-cohesive set of vertices, maximally l-cohesive subsets are recursively
    identified with l > k. Thus a hierarchy of vertex subsets is obtained in
    the end, with the entire graph G at its root.

    @return: an instance of L{CohesiveBlocks}. See the documentation of
      L{CohesiveBlocks} for more information.
    @see: L{CohesiveBlocks}
    """
    return CohesiveBlocks(graph, *GraphBase.cohesive_blocks(graph))


def _connected_components(graph, mode="strong"):
    """Calculates the (strong or weak) connected components for
    a given graph.

    @param mode: must be either C{"strong"} or C{"weak"}, depending on the
      connected components being sought. Optional, defaults to C{"strong"}.
    @return: a L{VertexClustering} object"""
    return VertexClustering(graph, GraphBase.connected_components(graph, mode))


def _clusters(graph, mode="strong"):
    """Deprecated alias to L{Graph.connected_components()}."""
    deprecated(
        "Graph.clusters() is deprecated; use Graph.connected_components() instead"
    )
    return graph.connected_components(mode=mode)



from igraph._igraph import GraphBase
from igraph.matching import Matching


def _maximum_bipartite_matching(graph, types="type", weights=None, eps=None):
    """Finds a maximum matching in a bipartite graph.

    A maximum matching is a set of edges such that each vertex is incident on
    at most one matched edge and the number (or weight) of such edges in the
    set is as large as possible.

    @param types: vertex types in a list or the name of a vertex attribute
      holding vertex types. Types should be denoted by zeros and ones (or
      C{False} and C{True}) for the two sides of the bipartite graph.
      If omitted, it defaults to C{type}, which is the default vertex type
      attribute for bipartite graphs.
    @param weights: edge weights to be used. Can be a sequence or iterable or
      even an edge attribute name.
    @param eps: a small real number used in equality tests in the weighted
      bipartite matching algorithm. Two real numbers are considered equal in
      the algorithm if their difference is smaller than this value. This
      is required to avoid the accumulation of numerical errors. If you
      pass C{None} here, igraph will try to determine an appropriate value
      automatically.
    @return: an instance of L{Matching}."""
    if eps is None:
        eps = -1

    matches = GraphBase._maximum_bipartite_matching(graph, types, weights, eps)
    return Matching(graph, matches, types=types)


def _bipartite_projection(
    graph, types="type", multiplicity=True, probe1=-1, which="both"
):
    """Projects a bipartite graph into two one-mode graphs. Edge directions
    are ignored while projecting.

    Examples:

    >>> g = Graph.Full_Bipartite(10, 5)
    >>> g1, g2 = g.bipartite_projection()
    >>> g1.isomorphic(Graph.Full(10))
    True
    >>> g2.isomorphic(Graph.Full(5))
    True

    @param types: an igraph vector containing the vertex types, or an
      attribute name. Anything that evalulates to C{False} corresponds to
      vertices of the first kind, everything else to the second kind.
    @param multiplicity: if C{True}, then igraph keeps the multiplicity of
      the edges in the projection in an edge attribute called C{"weight"}.
      E.g., if there is an A-C-B and an A-D-B triplet in the bipartite
      graph and there is no other X (apart from X=B and X=D) for which an
      A-X-B triplet would exist in the bipartite graph, the multiplicity
      of the A-B edge in the projection will be 2.
    @param probe1: this argument can be used to specify the order of the
      projections in the resulting list. If given and non-negative, then
      it is considered as a vertex ID; the projection containing the
      vertex will be the first one in the result.
    @param which: this argument can be used to specify which of the two
      projections should be returned if only one of them is needed. Passing
      0 here means that only the first projection is returned, while 1 means
      that only the second projection is returned. (Note that we use 0 and 1
      because Python indexing is zero-based). C{False} is equivalent to 0 and
      C{True} is equivalent to 1. Any other value means that both projections
      will be returned in a tuple.
    @return: a tuple containing the two projected one-mode graphs if C{which}
      is not 1 or 2, or the projected one-mode graph specified by the
      C{which} argument if its value is 0, 1, C{False} or C{True}.
    """
    # Deferred import to avoid cycles
    from igraph import Graph

    superclass_meth = super(Graph, graph).bipartite_projection

    if which is False:
        which = 0
    elif which is True:
        which = 1
    if which != 0 and which != 1:
        which = -1

    if multiplicity:
        if which == 0:
            g1, w1 = superclass_meth(types, True, probe1, which)
            g2, w2 = None, None
        elif which == 1:
            g1, w1 = None, None
            g2, w2 = superclass_meth(types, True, probe1, which)
        else:
            g1, g2, w1, w2 = superclass_meth(types, True, probe1, which)

        if g1 is not None:
            g1.es["weight"] = w1
            if g2 is not None:
                g2.es["weight"] = w2
                return g1, g2
            else:
                return g1
        else:
            g2.es["weight"] = w2
            return g2
    else:
        return superclass_meth(types, False, probe1, which)


def _bipartite_projection_size(graph, types="type", *args, **kwds):
    """Calculates the number of vertices and edges in the bipartite
    projections of this graph according to the specified vertex types.
    This is useful if you have a bipartite graph and you want to estimate
    the amount of memory you would need to calculate the projections
    themselves.

    @param types: an igraph vector containing the vertex types, or an
      attribute name. Anything that evalulates to C{False} corresponds to
      vertices of the first kind, everything else to the second kind.
    @return: a 4-tuple containing the number of vertices and edges in the
      first projection, followed by the number of vertices and edges in the
      second projection.
    """
    # Deferred import to avoid cycles
    from igraph import Graph

    return super(Graph, graph).bipartite_projection_size(types, *args, **kwds)



"""This module provides implementation for a Cairo-specific palette drawer"""

from igraph.drawing.cairo.base import AbstractCairoDrawer

__all__ = ("CairoPaletteDrawer",)


class CairoPaletteDrawer(AbstractCairoDrawer):
    """Default Cairo drawer object for palettes"""

    def __init__(self, context):
        """Constructs the vertex drawer and associates it to the given
        palette.

        @param context: the context on which we will draw
        """
        super().__init__(context, bbox=None)

    def draw(self, palette, **kwds):
        """TODO"""
        from igraph.datatypes import Matrix
        from igraph.drawing.utils import str_to_orientation

        context = self.context
        orientation = str_to_orientation(kwds.get("orientation", "lr"))

        # Construct a matrix and plot that
        indices = list(range(len(self)))
        if orientation in ("rl", "bt"):
            indices.reverse()
        if orientation in ("lr", "rl"):
            matrix = Matrix([indices])
        else:
            matrix = Matrix([[i] for i in indices])

        bbox = self.bbox = kwds.pop("bbox", None)
        if bbox is None:
            raise ValueError("bbox is required for Cairo plots")

        border_width = float(kwds.get("border_width", 1.0))
        grid_width = float(kwds.get("grid_width", 0.0))

        return matrix.__plot__(
            "cairo",
            context,
            bbox=bbox,
            palette=self,
            style="palette",
            square=False,
            grid_width=grid_width,
            border_width=border_width,
        )



def _export_graph_to_networkx(
    graph,
    create_using=None,
    vertex_attr_hashable: str = "_nx_name",
):
    """Converts the graph to networkx format.

    igraph has ordered vertices and edges, but networkx does not. To keep
    track of the original order, the '_igraph_index' vertex property is
    added to both vertices and edges.

    @param create_using: specifies which NetworkX graph class to use when
        constructing the graph. C{None} means to let igraph infer the most
        appropriate class based on whether the graph is directed and whether
        it has multi-edges.
    @param vertex_attr_hashable: vertex attribute used to name vertices
        in the exported network. The default "_nx_name" ensures round trip
        conversions to/from networkx are lossless.
    """
    import networkx as nx

    # Graph: decide on directness and mutliplicity
    if create_using is None:
        if graph.has_multiple():
            cls = nx.MultiDiGraph if graph.is_directed() else nx.MultiGraph
        else:
            cls = nx.DiGraph if graph.is_directed() else nx.Graph
    else:
        cls = create_using

    # Graph attributes
    kw = {x: graph[x] for x in graph.attributes()}
    g = cls(**kw)

    multigraph = isinstance(g, (nx.MultiGraph, nx.MultiDiGraph))

    # Nodes and node attributes
    for i, v in enumerate(graph.vs):
        vattr = v.attributes()
        vattr["_igraph_index"] = i

        # use _nx_name if the attribute is present so we can achieve
        # a lossless round-trip in terms of vertex names
        if vertex_attr_hashable in vattr:
            hashable = vattr.pop(vertex_attr_hashable)
        else:
            hashable = i

        # adding nodes one at a time is not slower in networkx
        g.add_node(hashable, **vattr)

    # Edges and edge attributes
    for i, edge in enumerate(graph.es):
        eattr = edge.attributes()
        eattr["_igraph_index"] = i

        if multigraph and "_nx_multiedge_key" in eattr:
            eattr["key"] = eattr.pop("_nx_multiedge_key")

        if vertex_attr_hashable in graph.vertex_attributes():
            hashable_source = graph.vs[vertex_attr_hashable][edge.source]
            hashable_target = graph.vs[vertex_attr_hashable][edge.target]
        else:
            hashable_source = edge.source
            hashable_target = edge.target

        # adding edges one at a time is not slower in networkx
        g.add_edge(hashable_source, hashable_target, **eattr)

    return g


def _construct_graph_from_networkx(cls, g, vertex_attr_hashable: str = "_nx_name"):
    """Converts the graph from networkx

    Vertex names will be stored as a vertex_attr_hashable attribute (usually
    "_nx_name", but see below). Because igraph stored vertices in an
    ordered manner, vertices will get new IDs from 0 up. In case of
    multigraphs, each edge will have an "_nx_multiedge_key" attribute, to
    distinguish edges that connect the same two vertices.

    @param g: networkx Graph or DiGraph
    @param vertex_attr_hashable: attribute used to store the Python
        hashable used by networkx to identify each vertex. The default value
        '_nx_name' ensures lossless round trip conversions to/from networkx. An
        alternative choice is 'name': in that case, using strings for vertex
        names is recommended and, if the graph is re-exported to networkx,
        Graph.to_networkx(vertex_attr_hashable="name") must be used to recover
        the correct vertex nomenclature in the exported network.

    """
    import networkx as nx

    # Graph attributes
    gattr = dict(g.graph)

    # Nodes
    vnames = list(g.nodes)
    vattr = {vertex_attr_hashable: vnames}
    vcount = len(vnames)

    # Dictionary connecting networkx hashables with igraph indices
    if len(g) and "_igraph_index" in next(iter(g.nodes.values())):
        # Collect _igraph_index and fill gaps
        idx = [x["_igraph_index"] for v, x in g.nodes.data()]
        idx.sort()
        idx_dict = {x: i for i, x in enumerate(idx)}

        vd = {}
        for v, datum in g.nodes.data():
            vd[v] = idx_dict[datum["_igraph_index"]]
    else:
        vd = {v: i for i, v in enumerate(vnames)}

    # NOTE: we do not need a special class for multigraphs, it is taken
    # care for at the edge level rather than at the graph level.
    graph = cls(
        n=vcount, directed=g.is_directed(), graph_attrs=gattr, vertex_attrs=vattr
    )

    # Vertex attributes
    for v, datum in g.nodes.data():
        for key, val in datum.items():
            # Get rid of _igraph_index (we used it already)
            if key == "_igraph_index":
                continue
            graph.vs[vd[v]][key] = val

    # Edges and edge attributes
    eattr_names = {name for (_, _, data) in g.edges.data() for name in data}
    eattr = {name: [] for name in eattr_names}
    edges = []
    # Multigraphs need a hidden attribute for multiedges
    if isinstance(g, (nx.MultiGraph, nx.MultiDiGraph)):
        eattr["_nx_multiedge_key"] = []
        for u, v, edgekey, data in g.edges.data(keys=True):
            edges.append((vd[u], vd[v]))
            for name in eattr_names:
                eattr[name].append(data.get(name))
            eattr["_nx_multiedge_key"].append(edgekey)

    else:
        for u, v, data in g.edges.data():
            edges.append((vd[u], vd[v]))
            for name in eattr_names:
                eattr[name].append(data.get(name))

    # Sort edges if there is a trace of a previous igraph ordering
    if "_igraph_index" in eattr:
        # Poor man's argsort
        sortd = list(enumerate(eattr["_igraph_index"]))
        sortd.sort(key=lambda x: x[1])
        idx = [i for i, x in sortd]

        # Get rid of the _igraph_index now
        del eattr["_igraph_index"]

        # Sort edges
        edges = [edges[i] for i in idx]
        # Sort each attribute
        eattr = {key: [val[i] for i in idx] for key, val in eattr.items()}

    graph.add_edges(edges, eattr)

    return graph


def _export_graph_to_graph_tool(
    graph, graph_attributes=None, vertex_attributes=None, edge_attributes=None
):
    """Converts the graph to graph-tool

    Data types: graph-tool only accepts specific data types. See the
    following web page for a list:

    https://graph-tool.skewed.de/static/doc/quickstart.html

    Note: because of the restricted data types in graph-tool, vertex and
    edge attributes require to be type-consistent across all vertices or
    edges. If you set the property for only some vertices/edges, the other
    will be tagged as None in igraph, so they can only be converted
    to graph-tool with the type 'object' and any other conversion will
    fail.

    @param graph_attributes: dictionary of graph attributes to transfer.
      Keys are attributes from the graph, values are data types (see
      below). C{None} means no graph attributes are transferred.
    @param vertex_attributes: dictionary of vertex attributes to transfer.
      Keys are attributes from the vertices, values are data types (see
      below). C{None} means no vertex attributes are transferred.
    @param edge_attributes: dictionary of edge attributes to transfer.
      Keys are attributes from the edges, values are data types (see
      below). C{None} means no vertex attributes are transferred.
    """
    import graph_tool as gt

    # Graph
    g = gt.Graph(directed=graph.is_directed())

    # Nodes
    vc = graph.vcount()
    g.add_vertex(vc)

    # Graph attributes
    if graph_attributes is not None:
        for x, dtype in graph_attributes.items():
            # Strange syntax for setting internal properties
            gprop = g.new_graph_property(str(dtype))
            g.graph_properties[x] = gprop
            g.graph_properties[x] = graph[x]

    # Vertex attributes
    if vertex_attributes is not None:
        for x, dtype in vertex_attributes.items():
            # Create a new vertex property
            g.vertex_properties[x] = g.new_vertex_property(str(dtype))
            # Fill the values from the igraph.Graph
            for i in range(vc):
                g.vertex_properties[x][g.vertex(i)] = graph.vs[i][x]

    # Edges and edge attributes
    if edge_attributes is not None:
        for x, dtype in edge_attributes.items():
            g.edge_properties[x] = g.new_edge_property(str(dtype))
    for edge in graph.es:
        e = g.add_edge(edge.source, edge.target)
        if edge_attributes is not None:
            for x in edge_attributes.keys():
                prop = edge.attributes().get(x, None)
                g.edge_properties[x][e] = prop

    return g


def _construct_graph_from_graph_tool(cls, g):
    """Converts the graph from graph-tool

    @param g: graph-tool Graph
    """
    # Graph attributes
    gattr = dict(g.graph_properties)

    # Nodes
    vcount = g.num_vertices()

    # Graph
    graph = cls(n=vcount, directed=g.is_directed(), graph_attrs=gattr)

    # Node attributes
    for key, val in g.vertex_properties.items():
        # val.get_array() returns None for non-scalar types so use the slower
        # way if this happens
        prop = val.get_array()
        arr = prop if prop is not None else val
        for i in range(vcount):
            graph.vs[i][key] = arr[i]

    # Edges and edge attributes
    # NOTE: graph-tool is quite strongly typed, so each property is always
    # defined for all edges, using default values for the type. E.g. for a
    # string property/attribute the missing edges get an empty string.
    edges = []
    eattr_names = list(g.edge_properties)
    eattr = {name: [] for name in eattr_names}
    for e in g.edges():
        edges.append((int(e.source()), int(e.target())))
        for name, attr_map in g.edge_properties.items():
            eattr[name].append(attr_map[e])

    graph.add_edges(edges, eattr)

    return graph



import unittest

from igraph import Graph

from .utils import temporary_file


class CliqueTests(unittest.TestCase):
    def setUp(self):
        self.g = Graph.Full(6)
        self.g.delete_edges([(0, 1), (0, 2), (3, 5)])

    def testCliques(self):
        tests = {
            (4, -1): [[1, 2, 3, 4], [1, 2, 4, 5]],
            (2, 2): [
                [0, 3],
                [0, 4],
                [0, 5],
                [1, 2],
                [1, 3],
                [1, 4],
                [1, 5],
                [2, 3],
                [2, 4],
                [2, 5],
                [3, 4],
                [4, 5],
            ],
            (-1, -1): [
                [0],
                [1],
                [2],
                [3],
                [4],
                [5],
                [0, 3],
                [0, 4],
                [0, 5],
                [1, 2],
                [1, 3],
                [1, 4],
                [1, 5],
                [2, 3],
                [2, 4],
                [2, 5],
                [3, 4],
                [4, 5],
                [0, 3, 4],
                [0, 4, 5],
                [1, 2, 3],
                [1, 2, 4],
                [1, 2, 5],
                [1, 3, 4],
                [1, 4, 5],
                [2, 3, 4],
                [2, 4, 5],
                [1, 2, 3, 4],
                [1, 2, 4, 5],
            ],
        }
        for (lo, hi), exp in tests.items():
            self.assertEqual(sorted(exp), sorted(map(sorted, self.g.cliques(lo, hi))))

    def testLargestCliques(self):
        self.assertEqual(
            sorted(map(sorted, self.g.largest_cliques())), [[1, 2, 3, 4], [1, 2, 4, 5]]
        )
        self.assertTrue(all(map(self.g.is_clique, self.g.largest_cliques())))

    def testMaximalCliques(self):
        self.assertEqual(
            sorted(map(sorted, self.g.maximal_cliques())),
            [[0, 3, 4], [0, 4, 5], [1, 2, 3, 4], [1, 2, 4, 5]],
        )
        self.assertTrue(all(map(self.g.is_clique, self.g.maximal_cliques())))
        self.assertEqual(
            sorted(map(sorted, self.g.maximal_cliques(min=4))),
            [[1, 2, 3, 4], [1, 2, 4, 5]],
        )
        self.assertEqual(
            sorted(map(sorted, self.g.maximal_cliques(max=3))), [[0, 3, 4], [0, 4, 5]]
        )

    def testMaximalCliquesFile(self):
        def read_cliques(fname):
            with open(fname) as fp:
                return sorted(sorted(int(item) for item in line.split()) for line in fp)

        with temporary_file() as fname:
            self.g.maximal_cliques(file=fname)
            self.assertEqual(
                [[0, 3, 4], [0, 4, 5], [1, 2, 3, 4], [1, 2, 4, 5]], read_cliques(fname)
            )

        with temporary_file() as fname:
            self.g.maximal_cliques(min=4, file=fname)
            self.assertEqual([[1, 2, 3, 4], [1, 2, 4, 5]], read_cliques(fname))

        with temporary_file() as fname:
            self.g.maximal_cliques(max=3, file=fname)
            self.assertEqual([[0, 3, 4], [0, 4, 5]], read_cliques(fname))

    def testCliqueNumber(self):
        self.assertEqual(self.g.clique_number(), 4)
        self.assertEqual(self.g.omega(), 4)


class IndependentVertexSetTests(unittest.TestCase):
    def setUp(self):
        self.g1 = Graph.Tree(5, 2, "undirected")
        self.g2 = Graph.Tree(10, 2, "undirected")

    def testIndependentVertexSets(self):
        tests = {
            (4, -1): [],
            (2, 2): [(0, 3), (0, 4), (1, 2), (2, 3), (2, 4), (3, 4)],
            (-1, -1): [
                (0,),
                (1,),
                (2,),
                (3,),
                (4,),
                (0, 3),
                (0, 4),
                (1, 2),
                (2, 3),
                (2, 4),
                (3, 4),
                (0, 3, 4),
                (2, 3, 4),
            ],
        }
        for (lo, hi), exp in tests.items():
            self.assertEqual(exp, self.g1.independent_vertex_sets(lo, hi))

    def testLargestIndependentVertexSets(self):
        self.assertEqual(
            self.g1.largest_independent_vertex_sets(), [(0, 3, 4), (2, 3, 4)]
        )
        self.assertTrue(all(map(self.g1.is_independent_vertex_set, self.g1.largest_independent_vertex_sets())))

    def testMaximalIndependentVertexSets(self):
        self.assertEqual(
            self.g2.maximal_independent_vertex_sets(),
            [
                (0, 3, 4, 5, 6),
                (0, 3, 5, 6, 9),
                (0, 4, 5, 6, 7, 8),
                (0, 5, 6, 7, 8, 9),
                (1, 2, 7, 8, 9),
                (1, 5, 6, 7, 8, 9),
                (2, 3, 4),
                (2, 3, 9),
                (2, 4, 7, 8),
            ],
        )
        self.assertTrue(all(map(self.g2.is_independent_vertex_set, self.g2.maximal_independent_vertex_sets())))

    def testIndependenceNumber(self):
        self.assertEqual(self.g2.independence_number(), 6)
        self.assertEqual(self.g2.alpha(), 6)


class CliqueBenchmark:
    """This is a benchmark, not a real test case. You can run it
    using:

    >>> from igraph.test.cliques import CliqueBenchmark
    >>> CliqueBenchmark().run()
    """

    def __init__(self):
        from time import time
        import gc

        self.time = time
        self.gc_collect = gc.collect

    def run(self):
        self.printIntro()
        self.testRandom()
        self.testMoonMoser()
        self.testGRG()

    def printIntro(self):
        print("n = number of vertices")
        print("#cliques = number of maximal cliques found")
        print("t1 = time required to determine the clique number")
        print("t2 = time required to determine and save all maximal cliques")
        print()

    def timeit(self, g):
        start = self.time()
        g.clique_number()
        mid = self.time()
        cl = g.maximal_cliques()
        end = self.time()
        self.gc_collect()
        return len(cl), mid - start, end - mid

    def testRandom(self):
        np = {
            100: [0.6, 0.7],
            300: [0.1, 0.2, 0.3, 0.4],
            500: [0.1, 0.2, 0.3],
            700: [0.1, 0.2],
            1000: [0.1, 0.2],
            10000: [0.001, 0.003, 0.005, 0.01, 0.02],
        }

        print()
        print("Erdos-Renyi random graphs")
        print("       n        p #cliques        t1        t2")
        for n in sorted(np.keys()):
            for p in np[n]:
                g = Graph.Erdos_Renyi(n, p)
                result = self.timeit(g)
                print("%8d %8.3f %8d %8.4fs %8.4fs" % tuple([n, p] + list(result)))

    def testMoonMoser(self):
        ns = [15, 27, 33]

        print()
        print("Moon-Moser graphs")
        print("       n exp_clqs #cliques        t1        t2")
        for n in ns:
            n3 = n // 3
            types = list(range(n3)) * 3
            el = [
                (i, j)
                for i in range(n)
                for j in range(i + 1, n)
                if types[i] != types[j]
            ]
            g = Graph(n, el)
            result = self.timeit(g)
            print(
                "%8d %8d %8d %8.4fs %8.4fs" % tuple([n, (3 ** (n / 3))] + list(result))
            )

    def testGRG(self):
        ns = [100, 1000, 5000, 10000, 25000, 50000]

        print()
        print("Geometric random graphs")
        print("       n        d #cliques        t1        t2")
        for n in ns:
            d = 2.0 / (n**0.5)
            g = Graph.GRG(n, d)
            result = self.timeit(g)
            print("%8d %8.3f %8d %8.4fs %8.4fs" % tuple([n, d] + list(result)))


def suite():
    clique_suite = unittest.defaultTestLoader.loadTestsFromTestCase(CliqueTests)
    indvset_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        IndependentVertexSetTests
    )
    return unittest.TestSuite([clique_suite, indvset_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



import unittest
from math import hypot
from igraph import Graph, Layout, BoundingBox, InternalError
from igraph import umap_compute_weights


class LayoutTests(unittest.TestCase):
    def testConstructor(self):
        layout = Layout([(0, 0, 1), (0, 1, 0), (1, 0, 0)])
        self.assertEqual(layout.dim, 3)
        layout = Layout([(0, 0, 1), (0, 1, 0), (1, 0, 0)], 3)
        self.assertEqual(layout.dim, 3)
        self.assertRaises(ValueError, Layout, [(0, 1), (1, 0)], 3)

    def testIndexing(self):
        layout = Layout([(0, 0, 1), (0, 1, 0), (1, 0, 0), (2, 1, 3)])
        self.assertEqual(len(layout), 4)
        self.assertEqual(layout[1], [0, 1, 0])
        self.assertEqual(layout[3], [2, 1, 3])

        row = layout[2]
        row[2] = 1
        self.assertEqual(layout[2], [1, 0, 1])

        del layout[1]
        self.assertEqual(len(layout), 3)

    def testScaling(self):
        layout = Layout([(0, 0, 1), (0, 1, 0), (1, 0, 0), (2, 1, 3)])
        layout.scale(1.5)
        self.assertEqual(
            layout.coords,
            [[0.0, 0.0, 1.5], [0.0, 1.5, 0.0], [1.5, 0.0, 0.0], [3.0, 1.5, 4.5]],
        )

        layout = Layout([(0, 0, 1), (0, 1, 0), (1, 0, 0), (2, 1, 3)])
        layout.scale(1, 1, 3)
        self.assertEqual(layout.coords, [[0, 0, 3], [0, 1, 0], [1, 0, 0], [2, 1, 9]])

        layout = Layout([(0, 0, 1), (0, 1, 0), (1, 0, 0), (2, 1, 3)])
        layout.scale((2, 2, 1))
        self.assertEqual(layout.coords, [[0, 0, 1], [0, 2, 0], [2, 0, 0], [4, 2, 3]])

        self.assertRaises(ValueError, layout.scale, 2, 3)

    def testTranslation(self):
        layout = Layout([(0, 0, 1), (0, 1, 0), (1, 0, 0), (2, 1, 3)])
        layout2 = layout.copy()

        layout.translate(1, 3, 2)
        self.assertEqual(layout.coords, [[1, 3, 3], [1, 4, 2], [2, 3, 2], [3, 4, 5]])

        layout.translate((-1, -3, -2))
        self.assertEqual(layout.coords, layout2.coords)

        self.assertRaises(ValueError, layout.translate, v=[3])

    def testCentroid(self):
        layout = Layout([(0, 0, 1), (0, 1, 0), (1, 0, 0), (2, 1, 3)])
        centroid = layout.centroid()
        self.assertEqual(len(centroid), 3)
        self.assertAlmostEqual(centroid[0], 0.75)
        self.assertAlmostEqual(centroid[1], 0.5)
        self.assertAlmostEqual(centroid[2], 1.0)

    def testBoundaries(self):
        layout = Layout([(0, 0, 1), (0, 1, 0), (1, 0, 0), (2, 1, 3)])
        self.assertEqual(layout.boundaries(), ([0, 0, 0], [2, 1, 3]))
        self.assertEqual(layout.boundaries(1), ([-1, -1, -1], [3, 2, 4]))

        layout = Layout([])
        self.assertRaises(ValueError, layout.boundaries)
        layout = Layout([], dim=3)
        self.assertRaises(ValueError, layout.boundaries)

    def testBoundingBox(self):
        layout = Layout([(0, 1), (2, 7)])
        self.assertEqual(layout.bounding_box(), BoundingBox(0, 1, 2, 7))
        self.assertEqual(layout.bounding_box(1), BoundingBox(-1, 0, 3, 8))
        layout = Layout([])
        self.assertEqual(layout.bounding_box(), BoundingBox(0, 0, 0, 0))

    def testCenter(self):
        layout = Layout([(-2, 0), (-2, -2), (0, -2), (0, 0)])
        layout.center()
        self.assertEqual(layout.coords, [[-1, 1], [-1, -1], [1, -1], [1, 1]])
        layout.center(5, 5)
        self.assertEqual(layout.coords, [[4, 6], [4, 4], [6, 4], [6, 6]])
        self.assertRaises(ValueError, layout.center, 3)
        self.assertRaises(TypeError, layout.center, p=6)

    def testFitInto(self):
        layout = Layout([(-2, 0), (-2, -2), (0, -2), (0, 0)])
        layout.fit_into(BoundingBox(5, 5, 8, 10), keep_aspect_ratio=False)
        self.assertEqual(layout.coords, [[5, 10], [5, 5], [8, 5], [8, 10]])
        layout = Layout([(-2, 0), (-2, -2), (0, -2), (0, 0)])
        layout.fit_into(BoundingBox(5, 5, 8, 10))
        self.assertEqual(layout.coords, [[5, 9], [5, 6], [8, 6], [8, 9]])

        layout = Layout([(-1, -1, -1), (0, 0, 0), (1, 1, 1), (2, 2, 0), (3, 3, -1)])
        layout.fit_into((0, 0, 0, 8, 8, 4))
        self.assertEqual(
            layout.coords, [[0, 0, 0], [2, 2, 2], [4, 4, 4], [6, 6, 2], [8, 8, 0]]
        )

        layout = Layout([])
        layout.fit_into((6, 7, 8, 11))
        self.assertEqual(layout.coords, [])

    def testToPolar(self):
        layout = Layout([(0, 0), (-1, 1), (0, 1), (1, 1)])
        layout.to_radial(min_angle=180, max_angle=0, max_radius=2)
        exp = [[0.0, 0.0], [-2.0, 0.0], [0.0, 2.0], [2, 0.0]]
        for idx in range(4):
            self.assertAlmostEqual(layout.coords[idx][0], exp[idx][0], places=3)
            self.assertAlmostEqual(layout.coords[idx][1], exp[idx][1], places=3)

    def testTransform(self):
        def tr(coord, dx, dy):
            return coord[0] + dx, coord[1] + dy

        layout = Layout([(1, 2), (3, 4)])
        layout.transform(tr, 2, -1)
        self.assertEqual(layout.coords, [[3, 1], [5, 3]])


class LayoutAlgorithmTests(unittest.TestCase):
    def testAuto(self):
        def layout_test(graph, test_with_dims=(2, 3)):
            lo = graph.layout("auto")
            self.assertTrue(isinstance(lo, Layout))
            self.assertEqual(len(lo[0]), 2)
            for dim in test_with_dims:
                lo = graph.layout("auto", dim=dim)
                self.assertTrue(isinstance(lo, Layout))
                self.assertEqual(len(lo[0]), dim)
            return lo

        g = Graph.Barabasi(10)
        layout_test(g)

        g = Graph.GRG(101, 0.2)
        del g.vs["x"]
        del g.vs["y"]
        layout_test(g)

        g = Graph.Full(10) * 2
        layout_test(g)

        g["layout"] = "graphopt"
        layout_test(g, test_with_dims=())

        g.vs["x"] = list(range(20))
        g.vs["y"] = list(range(20, 40))
        layout_test(g, test_with_dims=())

        del g["layout"]
        lo = layout_test(g, test_with_dims=(2,))
        self.assertEqual(
            [tuple(item) for item in lo],
            list(zip(list(range(20)), list(range(20, 40)))),
        )

        g.vs["z"] = list(range(40, 60))
        lo = layout_test(g)
        self.assertEqual(
            [tuple(item) for item in lo],
            list(zip(list(range(20)), list(range(20, 40)), list(range(40, 60)))),
        )

    def testCircle(self):
        def test_is_proper_circular_layout(graph, layout):
            xs, ys = list(zip(*layout))
            n = graph.vcount()
            self.assertEqual(n, len(xs))
            self.assertEqual(n, len(ys))
            self.assertAlmostEqual(0, sum(xs))
            self.assertAlmostEqual(0, sum(ys))
            for x, y in zip(xs, ys):
                self.assertAlmostEqual(1, x**2 + y**2)

        g = Graph.Ring(8)
        layout = g.layout("circle")
        test_is_proper_circular_layout(g, g.layout("circle"))

        order = [0, 2, 4, 6, 1, 3, 5, 7]
        ordered_layout = g.layout("circle", order=order)
        test_is_proper_circular_layout(g, g.layout("circle"))
        for v, w in enumerate(order):
            self.assertAlmostEqual(layout[v][0], ordered_layout[w][0])
            self.assertAlmostEqual(layout[v][1], ordered_layout[w][1])

    def testDavidsonHarel(self):
        # Quick smoke testing only
        g = Graph.Barabasi(100)
        lo = g.layout("dh")
        self.assertTrue(isinstance(lo, Layout))

    def testFruchtermanReingold(self):
        g = Graph.Barabasi(100)

        lo = g.layout("fr")
        self.assertTrue(isinstance(lo, Layout))

        lo = g.layout("fr", miny=list(range(100)))
        self.assertTrue(isinstance(lo, Layout))
        self.assertTrue(all(lo[i][1] >= i for i in range(100)))

        lo = g.layout("fr", miny=list(range(100)), maxy=list(range(100)))
        self.assertTrue(isinstance(lo, Layout))
        self.assertTrue(all(lo[i][1] == i for i in range(100)))

        lo = g.layout(
            "fr", miny=[2] * 100, maxy=[3] * 100, minx=[4] * 100, maxx=[6] * 100
        )
        self.assertTrue(isinstance(lo, Layout))
        bbox = lo.bounding_box()
        self.assertTrue(bbox.top >= 2)
        self.assertTrue(bbox.bottom <= 3)
        self.assertTrue(bbox.left >= 4)
        self.assertTrue(bbox.right <= 6)

    def testFruchtermanReingoldGrid(self):
        g = Graph.Barabasi(100)
        for grid_opt in ["grid", "nogrid", "auto", True, False]:
            lo = g.layout("fr", miny=list(range(100)), grid=grid_opt)
            self.assertTrue(isinstance(lo, Layout))
            self.assertTrue(all(lo[i][1] >= i for i in range(100)))

    def testKamadaKawai(self):
        g = Graph.Barabasi(100)

        lo = g.layout(
            "kk", miny=[2] * 100, maxy=[3] * 100, minx=[4] * 100, maxx=[6] * 100
        )

        self.assertTrue(isinstance(lo, Layout))
        bbox = lo.bounding_box()
        self.assertTrue(bbox.top >= 2)
        self.assertTrue(bbox.bottom <= 3)
        self.assertTrue(bbox.left >= 4)
        self.assertTrue(bbox.right <= 6)

        lo = g.layout(
            "kk", miny=[2] * 100, maxy=[3] * 100, minx=[4] * 100, maxx=[6] * 100,
            weights=range(10, g.ecount() + 10)
        )

        self.assertTrue(isinstance(lo, Layout))
        bbox = lo.bounding_box()
        self.assertTrue(bbox.top >= 2)
        self.assertTrue(bbox.bottom <= 3)
        self.assertTrue(bbox.left >= 4)
        self.assertTrue(bbox.right <= 6)

    def testMDS(self):
        g = Graph.Tree(10, 2)
        lo = g.layout("mds")
        self.assertTrue(isinstance(lo, Layout))

        dists = g.distances()
        lo = g.layout("mds", dists)
        self.assertTrue(isinstance(lo, Layout))

        g += Graph.Tree(10, 2)
        lo = g.layout("mds")
        self.assertTrue(isinstance(lo, Layout))

    def testUMAP(self):
        g = Graph()

        self.assertRaises(
            InternalError,
            g.layout_umap,
            min_dist=-0.01,
        )

        self.assertRaises(
            ValueError,
            g.layout_umap,
            epochs=-1,
        )

        self.assertRaises(
            ValueError,
            g.layout_umap,
            dim=1,
        )

        # Empty graph
        lo = g.layout_umap()
        self.assertTrue(isinstance(lo, Layout))
        self.assertEqual(lo.coords, [])

        # Singleton graph
        g = Graph(n=1)
        lo = g.layout_umap()
        self.assertEqual(lo.coords, [[0, 0]])

        # Graph with two articulation points
        edges = [
            0,
            1,
            0,
            2,
            0,
            3,
            1,
            2,
            1,
            3,
            2,
            3,
            3,
            4,
            4,
            5,
            5,
            6,
            6,
            7,
            7,
            8,
            6,
            8,
            7,
            9,
            6,
            9,
            8,
            9,
            7,
            10,
            8,
            10,
            9,
            10,
            10,
            11,
            9,
            11,
            8,
            11,
            7,
            11,
        ]
        edges = list(zip(edges[::2], edges[1::2]))
        dist = [
            0.1,
            0.09,
            0.12,
            0.09,
            0.1,
            0.1,
            0.9,
            0.9,
            0.9,
            0.2,
            0.1,
            0.1,
            0.1,
            0.1,
            0.1,
            0.08,
            0.05,
            0.1,
            0.08,
            0.12,
            0.09,
            0.11,
        ]
        g = Graph(edges)
        lo = g.layout_umap(dist=dist, epochs=500)
        self.assertTrue(isinstance(lo, Layout))

        # One should get two clusters in this case
        x, y = list(zip(*lo.coords))
        xmax, ymax, xmin, ymin = max(x), max(y), min(x), min(y)
        distmax = max(xmax - xmin, ymax - ymin)
        for iclu in range(0, 8, 7):
            xclu = sum(x[iclu : iclu + 4]) / 4
            yclu = sum(y[iclu : iclu + 4]) / 4
            for i in range(4):
                dx = x[iclu + i] - xclu
                dy = y[iclu + i] - yclu
                dxy = hypot(dx, dy)
                # Distance from each cluster's center should be relatively small
                self.assertLess(dxy, 0.2 * distmax)

        # Test single epoch with seed
        lo_adj = g.layout_umap(dist=dist, epochs=1, seed=lo)
        self.assertTrue(isinstance(lo_adj, Layout))

        # Same but inputting the coordinates
        lo_adj = g.layout_umap(dist=dist, epochs=1, seed=lo.coords)
        self.assertTrue(isinstance(lo_adj, Layout))

    def testUMAPComputeWeights(self):
        edges = [0, 1, 0, 2, 1, 2, 1, 3, 2, 3, 2, 0]
        edges = list(zip(edges[::2], edges[1::2]))
        dist = [1, 1.5, 1.8, 2.0, 3.4, 0.5]
        # NOTE: you need a directed graph to make sense of the symmetryzation
        g = Graph(edges, directed=True)
        weights = umap_compute_weights(g, dist)
        self.assertEqual(
            weights, [1.0, 1.0, 1.0, 1.1253517471925912e-07, 6.14421235332821e-06, 0.0]
        )

    def testLGL(self):
        g = Graph.Barabasi(100)
        lo = g.layout("lgl")
        self.assertTrue(isinstance(lo, Layout))

        lo = g.layout("lgl", root=5)
        self.assertTrue(isinstance(lo, Layout))

    def testReingoldTilford(self):
        g = Graph.Barabasi(100)
        lo = g.layout("rt")
        ys = [coord[1] for coord in lo]
        root = ys.index(0.0)
        self.assertEqual(ys, g.distances(root)[0])
        g = Graph.Barabasi(100) + Graph.Barabasi(50)
        lo = g.layout("rt", root=[0, 100])
        self.assertEqual(lo[100][1] - lo[0][1], 0)
        lo = g.layout("rt", root=[0, 100], rootlevel=[2, 10])
        self.assertEqual(lo[100][1] - lo[0][1], 8)

        # test named vertices
        g.vs["name"] = [f"v{i}" for i in range(g.vcount())]
        lo = g.layout("rt", root=["v0", "v100"])
        self.assertEqual(lo[100][1] - lo[0][1], 0)

    def testBipartite(self):
        g = Graph.Full_Bipartite(3, 2)

        lo = g.layout("bipartite")
        ys = [coord[1] for coord in lo]
        self.assertEqual([1, 1, 1, 0, 0], ys)

        lo = g.layout("bipartite", vgap=3)
        ys = [coord[1] for coord in lo]
        self.assertEqual([3, 3, 3, 0, 0], ys)

        lo = g.layout("bipartite", hgap=5)
        self.assertEqual({0, 5, 10}, {coord[0] for coord in lo if coord[1] == 1})
        self.assertEqual({2.5, 7.5}, {coord[0] for coord in lo if coord[1] == 0})

    def testDRL(self):
        # Regression test for bug #1091891
        g = Graph.Ring(10, circular=False) + 1
        lo = g.layout("drl")
        self.assertTrue(isinstance(lo, Layout))


def suite():
    layout_suite = unittest.defaultTestLoader.loadTestsFromTestCase(LayoutTests)
    layout_algorithm_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        LayoutAlgorithmTests
    )
    return unittest.TestSuite([layout_suite, layout_algorithm_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



"""
========
Simplify
========

This example shows how to remove self loops and multiple edges using :meth:`igraph.GraphBase.simplify`.
"""
import igraph as ig
import matplotlib.pyplot as plt

# %%
# We start with a graph that includes loops and multiedges:
g1 = ig.Graph([
    (0, 1),
    (1, 2),
    (2, 3),
    (3, 4),
    (4, 0),
    (0, 0),
    (1, 4),
    (1, 4),
    (0, 2),
    (2, 4),
    (2, 4),
    (2, 4),
    (3, 3)],
)

# %%
# To simplify the graph, we must remember that the function operates in place,
# i.e. directly changes the graph that it is run on. So we need to first make a
# copy of our graph, and then simplify that copy to keep the original graph
# untouched:
g2 = g1.copy()
g2.simplify()

# %%
# We can then proceed to plot both graphs to see the difference. First, let's
# choose a consistent visual style:
visual_style = {
    "vertex_color": "lightblue",
    "vertex_size": 20,
    "vertex_label": [0, 1, 2, 3, 4],
}

# %%
# And finally, let's plot them in twin axes, with rectangular frames around
# each plot:
fig, axs = plt.subplots(1, 2, sharex=True, sharey=True)
ig.plot(
    g1,
    layout="circle",
    target=axs[0],
    **visual_style,
)
ig.plot(
    g2,
    layout="circle",
    target=axs[1],
    **visual_style,
)
axs[0].set_title('Multigraph...')
axs[1].set_title('...simplified')
# Draw rectangles around axes
axs[0].add_patch(plt.Rectangle(
    (0, 0), 1, 1, fc='none', ec='k', lw=4, transform=axs[0].transAxes,
    ))
axs[1].add_patch(plt.Rectangle(
    (0, 0), 1, 1, fc='none', ec='k', lw=4, transform=axs[1].transAxes,
    ))
plt.show()



"""This module provides implementations of Matplotlib-specific vertex drawers,
i.e. drawers that the Matplotlib graph drawer will use to draw vertices.
"""

from math import pi

from igraph.drawing.baseclasses import AbstractVertexDrawer
from igraph.drawing.metamagic import AttributeCollectorBase
from igraph.drawing.shapes import ShapeDrawerDirectory
from igraph.drawing.matplotlib.utils import find_matplotlib
from igraph.drawing.utils import FakeModule

mpl, _ = find_matplotlib()
try:
    IdentityTransform = mpl.transforms.IdentityTransform
    PatchCollection = mpl.collections.PatchCollection
except AttributeError:
    IdentityTransform = FakeModule
    PatchCollection = FakeModule


__all__ = ("MatplotlibVertexDrawer", "VertexCollection")


class MatplotlibVertexDrawer(AbstractVertexDrawer):
    """Matplotlib backend-specific vertex drawer."""

    def __init__(self, ax, palette, layout):
        self.context = ax
        super().__init__(palette, layout)
        self.VisualVertexBuilder = self._construct_visual_vertex_builder()

    def _construct_visual_vertex_builder(self):
        class VisualVertexBuilder(AttributeCollectorBase):
            """Collects some visual properties of a vertex for drawing"""

            _kwds_prefix = "vertex_"
            color = ("red", self.palette.get)
            frame_color = ("black", self.palette.get)
            frame_width = 1.0
            label = None
            label_angle = -pi / 2
            label_dist = 0.0
            label_color = ("black", self.palette.get)
            font = "sans-serif"
            label_size = 12.0
            # FIXME? mpl.rcParams["font.size"])
            position = {"func": self.layout.__getitem__}
            shape = ("circle", ShapeDrawerDirectory.resolve_default)
            size = 30
            width = None
            height = None
            zorder = 2

        return VisualVertexBuilder

    def draw(self, visual_vertex, vertex, coords):
        """Build the Artist for a vertex and return it."""
        ax = self.context

        width = (
            visual_vertex.width
            if visual_vertex.width is not None
            else visual_vertex.size
        )
        height = (
            visual_vertex.height
            if visual_vertex.height is not None
            else visual_vertex.size
        )

        art = visual_vertex.shape.draw_path(
            ax,
            0,
            0,
            width,
            height,
            facecolor=visual_vertex.color,
            edgecolor=visual_vertex.frame_color,
            linewidth=visual_vertex.frame_width,
            zorder=visual_vertex.zorder,
            transform=IdentityTransform(),
        )
        return art


class VertexCollection(PatchCollection):
    """Collection of vertex patches for plotting.

    This class takes additional keyword arguments compared to PatchCollection:

    @param vertex_builder: A list of vertex builders to construct the visual
        vertices. This is updated if the size of the vertices is changed.
    @param size_callback: A function to be triggered after vertex sizes are
        changed. Typically this redraws the edges.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._stale_size = False

    def get_sizes(self):
        """Same as get_size."""
        return self.get_size()

    def get_size(self):
        """Get vertex sizes.

        If width and height are unequal, get the largest of the two.

        @return: An array of vertex sizes.
        """
        import numpy as np

        sizes = []
        for path in self.get_paths():
            bbox = path.get_extents()
            mins, maxs = bbox.min, bbox.max
            width, height = maxs - mins
            size = max(width, height)
            sizes.append(size)
        return np.array(sizes)

    def set_size(self, sizes):
        """Set vertex sizes.

        This rescales the current vertex symbol/path linearly, using this
        value as the largest of width and height.

        @param sizes: A sequence of vertex sizes or a single size.
        """
        paths = self._paths
        try:
            iter(sizes)
        except TypeError:
            sizes = [sizes] * len(paths)

        sizes = list(sizes)
        current_sizes = self.get_sizes()
        for path, cursize in zip(paths, current_sizes):
            # Circular use of sizes
            size = sizes.pop(0)
            sizes.append(size)
            # Rescale the path for this vertex
            path.vertices *= size / cursize

        self._stale_size = True
        self.stale = True

    def set_sizes(self, sizes):
        """Same as set_size."""
        self.set_size(sizes)

    @property
    def stale(self):
        return super().stale

    @stale.setter
    def stale(self, val):
        PatchCollection.stale.fset(self, val)
        if val and hasattr(self, "stale_callback_post"):
            self.stale_callback_post(self)
